
"""
List of elements

Data processed from wikipedia's periodic table template page (fall 2014)
"""

elements = [[
    {'Z': 1, 'name': "hydrogen", 'id': "H", 'phase': "Gas  ",
        'category': "Diatomic nonmetal", 'occurrence': "Primordial"},
    (17, ""),
    {'Z': 2, 'name': "helium", 'id': "He", 'phase': "Gas  ",
        'category': "Noble gas", 'occurrence': "Primordial"},

], [
    {'Z': 3, 'name': "lithium", 'id': "Li", 'phase': "Solid",
        'category': "Alkali metal", 'occurrence': "Primordial"},
    {'Z': 4, 'name': "beryllium", 'id': "Be", 'phase': "Solid",
        'category': "Alkaline earth metal", 'occurrence': "Primordial"},
    (11, ""),
    {'Z': 5, 'name': "boron", 'id': "B", 'phase': "Solid",
        'category': "Metalloid", 'occurrence': "Primordial"},
    {'Z': 6, 'name': "carbon", 'id': "C", 'phase': "Solid",
        'category': "Polyatomic nonmetal", 'occurrence': "Primordial"},
    {'Z': 7, 'name': "nitrogen", 'id': "N", 'phase': "Gas  ",
        'category': "Diatomic nonmetal", 'occurrence': "Primordial"},
    {'Z': 8, 'name': "oxygen", 'id': "O", 'phase': "Gas  ",
        'category': "Diatomic nonmetal", 'occurrence': "Primordial"},
    {'Z': 9, 'name': "fluorine", 'id': "F", 'phase': "Gas  ",
        'category': "Diatomic nonmetal", 'occurrence': "Primordial"},
    {'Z': 10, 'name': "neon", 'id': "Ne", 'phase': "Gas  ",
        'category': "Noble gas", 'occurrence': "Primordial"},

], [
    {'Z': 11, 'name': "sodium", 'id': "Na", 'phase': "Solid",
        'category': "Alkali metal", 'occurrence': "Primordial"},
    {'Z': 12, 'name': "magnesium", 'id': "Mg", 'phase': "Solid",
        'category': "Alkaline earth metal", 'occurrence': "Primordial"},
    (11, ""),
    {'Z': 13, 'name': "aluminium", 'id': "Al", 'phase': "Solid",
        'category': "post-transition metal", 'occurrence': "Primordial"},
    {'Z': 14, 'name': "silicon", 'id': "Si", 'phase': "Solid",
        'category': "Metalloid", 'occurrence': "Primordial"},
    {'Z': 15, 'name': "phosphorus", 'id': "P", 'phase': "Solid",
        'category': "Polyatomic nonmetal", 'occurrence': "Primordial"},
    {'Z': 16, 'name': "sulfur", 'id': "S", 'phase': "Solid",
        'category': "Polyatomic nonmetal", 'occurrence': "Primordial"},
    {'Z': 17, 'name': "chlorine", 'id': "Cl", 'phase': "Gas  ",
        'category': "Diatomic nonmetal", 'occurrence': "Primordial"},
    {'Z': 18, 'name': "argon", 'id': "Ar", 'phase': "Gas  ",
        'category': "Noble gas", 'occurrence': "Primordial"},

], [
    {'Z': 19, 'name': "potassium", 'id': "K", 'phase': "Solid",
        'category': "Alkali metal", 'occurrence': "Primordial"},
    {'Z': 20, 'name': "calcium", 'id': "Ca", 'phase': "Solid",
        'category': "Alkaline earth metal", 'occurrence': "Primordial"},
    (1, ""),
    {'Z': 21, 'name': "scandium", 'id': "Sc", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 22, 'name': "titanium", 'id': "Ti", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 23, 'name': "vanadium", 'id': "V", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 24, 'name': "chromium", 'id': "Cr", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 25, 'name': "manganese", 'id': "Mn", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 26, 'name': "iron", 'id': "Fe", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 27, 'name': "cobalt", 'id': "Co", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 28, 'name': "nickel", 'id': "Ni", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 29, 'name': "copper", 'id': "Cu", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 30, 'name': "zinc", 'id': "Zn", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 31, 'name': "gallium", 'id': "Ga", 'phase': "Solid",
        'category': "post-transition metal", 'occurrence': "Primordial"},
    {'Z': 32, 'name': "germanium", 'id': "Ge", 'phase': "Solid",
        'category': "Metalloid", 'occurrence': "Primordial"},
    {'Z': 33, 'name': "arsenic", 'id': "As", 'phase': "Solid",
        'category': "Metalloid", 'occurrence': "Primordial"},
    {'Z': 34, 'name': "selenium", 'id': "Se", 'phase': "Solid",
        'category': "Polyatomic nonmetal", 'occurrence': "Primordial"},
    {'Z': 35, 'name': "bromine", 'id': "Br", 'phase': "Liquid",
        'category': "Diatomic nonmetal", 'occurrence': "Primordial"},
    {'Z': 36, 'name': "krypton", 'id': "Kr", 'phase': "Gas  ",
        'category': "Noble gas", 'occurrence': "Primordial"},

], [
    {'Z': 37, 'name': "rubidium", 'id': "Rb", 'phase': "Solid",
        'category': "Alkali metal", 'occurrence': "Primordial"},
    {'Z': 38, 'name': "strontium", 'id': "Sr", 'phase': "Solid",
        'category': "Alkaline earth metal", 'occurrence': "Primordial"},
    (1, ""),
    {'Z': 39, 'name': "yttrium", 'id': "Y", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 40, 'name': "zirconium", 'id': "Zr", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 41, 'name': "niobium", 'id': "Nb", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 42, 'name': "molybdenum", 'id': "Mo", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 43, 'name': "technetium", 'id': "Tc", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "From decay"},
    {'Z': 44, 'name': "ruthenium", 'id': "Ru", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 45, 'name': "rhodium", 'id': "Rh", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 46, 'name': "palladium", 'id': "Pd", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 47, 'name': "silver", 'id': "Ag", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 48, 'name': "cadmium", 'id': "Cd", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 49, 'name': "indium", 'id': "In", 'phase': "Solid",
        'category': "post-transition metal", 'occurrence': "Primordial"},
    {'Z': 50, 'name': "tin", 'id': "Sn", 'phase': "Solid",
        'category': "post-transition metal", 'occurrence': "Primordial"},
    {'Z': 51, 'name': "antimony", 'id': "Sb", 'phase': "Solid",
        'category': "Metalloid", 'occurrence': "Primordial"},
    {'Z': 52, 'name': "tellurium", 'id': "Te", 'phase': "Solid",
        'category': "Metalloid", 'occurrence': "Primordial"},
    {'Z': 53, 'name': "iodine", 'id': "I", 'phase': "Solid",
        'category': "Diatomic nonmetal", 'occurrence': "Primordial"},
    {'Z': 54, 'name': "xenon", 'id': "Xe", 'phase': "Gas  ",
        'category': "Noble gas", 'occurrence': "Primordial"},

], [
    {'Z': 55, 'name': "caesium", 'id': "Cs", 'phase': "Solid",
        'category': "Alkali metal", 'occurrence': "Primordial"},
    {'Z': 56, 'name': "barium", 'id': "Ba", 'phase': "Solid",
        'category': "Alkaline earth metal", 'occurrence': "Primordial"},
    (1, "*"),
    {'Z': 71, 'name': "lutetium", 'id': "Lu", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 72, 'name': "hafnium", 'id': "Hf", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 73, 'name': "tantalum", 'id': "Ta", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 74, 'name': "tungsten", 'id': "W", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 75, 'name': "rhenium", 'id': "Re", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 76, 'name': "osmium", 'id': "Os", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 77, 'name': "iridium", 'id': "Ir", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 78, 'name': "platinum", 'id': "Pt", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 79, 'name': "gold", 'id': "Au", 'phase': "Solid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 80, 'name': "mercury", 'id': "Hg", 'phase': "Liquid",
        'category': "Transition metal", 'occurrence': "Primordial"},
    {'Z': 81, 'name': "thallium", 'id': "Tl", 'phase': "Solid",
        'category': "post-transition metal", 'occurrence': "Primordial"},
    {'Z': 82, 'name': "lead", 'id': "Pb", 'phase': "Solid",
        'category': "post-transition metal", 'occurrence': "Primordial"},
    {'Z': 83, 'name': "bismuth", 'id': "Bi", 'phase': "Solid",
        'category': "post-transition metal", 'occurrence': "Primordial"},
    {'Z': 84, 'name': "polonium", 'id': "Po", 'phase': "Solid",
        'category': "post-transition metal", 'occurrence': "From decay"},
    {'Z': 85, 'name': "astatine", 'id': "At", 'phase': "Solid",
        'category': "Metalloid", 'occurrence': "From decay"},
    {'Z': 86, 'name': "radon", 'id': "Rn", 'phase': "Gas  ",
        'category': "Noble gas", 'occurrence': "From decay"},

], [
    {'Z': 87, 'name': "francium", 'id': "Fr", 'phase': "Solid",
        'category': "Alkali metal", 'occurrence': "From decay"},
    {'Z': 88, 'name': "radium", 'id': "Ra", 'phase': "Solid",
        'category': "Alkaline earth metal", 'occurrence': "From decay"},
    (1, "**"),
    {'Z': 103, 'name': "lawrencium", 'id': "Lr", 'phase': "Unknown phase",
        'category': "Actinide", 'occurrence': "Synthetic"},
    {'Z': 104, 'name': "rutherfordium", 'id': "Rf", 'phase': "Unknown phase",
        'category': "Transition metal", 'occurrence': "Synthetic"},
    {'Z': 105, 'name': "dubnium", 'id': "Db", 'phase': "Unknown phase",
        'category': "Transition metal", 'occurrence': "Synthetic"},
    {'Z': 106, 'name': "seaborgium", 'id': "Sg", 'phase': "Unknown phase",
        'category': "Transition metal", 'occurrence': "Synthetic"},
    {'Z': 107, 'name': "bohrium", 'id': "Bh", 'phase': "Unknown phase",
        'category': "Transition metal", 'occurrence': "Synthetic"},
    {'Z': 108, 'name': "hassium", 'id': "Hs", 'phase': "Unknown phase",
        'category': "Transition metal", 'occurrence': "Synthetic"},
    {'Z': 109, 'name': "meitnerium", 'id': "Mt", 'phase': "Unknown phase",
        'category': "Unknown chemical properties", 'occurrence': "Synthetic"},
    {'Z': 110, 'name': "darmstadtium", 'id': "Ds", 'phase': "Unknown phase",
        'category': "Unknown chemical properties", 'occurrence': "Synthetic"},
    {'Z': 111, 'name': "roentgenium", 'id': "Rg", 'phase': "Unknown phase",
        'category': "Unknown chemical properties", 'occurrence': "Synthetic"},
    {'Z': 112, 'name': "copernicium", 'id': "Cn", 'phase': "Unknown phase",
        'category': "Transition metal", 'occurrence': "Synthetic"},
    {'Z': 113, 'name': "ununtrium", 'id': "Uut", 'phase': "Unknown phase",
        'category': "Unknown chemical properties", 'occurrence': "Synthetic"},
    {'Z': 114, 'name': "flerovium", 'id': "Fl", 'phase': "Unknown phase",
        'category': "Post-transition metal", 'occurrence': "Synthetic"},
    {'Z': 115, 'name': "ununpentium", 'id': "Uup", 'phase': "Unknown phase",
        'category': "Unknown chemical properties", 'occurrence': "Synthetic"},
    {'Z': 116, 'name': "livermorium", 'id': "Lv", 'phase': "Unknown phase",
        'category': "Unknown chemical properties", 'occurrence': "Synthetic"},
    {'Z': 117, 'name': "ununseptium", 'id': "Uus", 'phase': "Unknown phase",
        'category': "Unknown chemical properties", 'occurrence': "Synthetic"},
    {'Z': 118, 'name': "ununoctium", 'id': "Uuo", 'phase': "Unknown phase",
        'category': "Unknown chemical properties", 'occurrence': "Synthetic"},

], [
    (3, "*"),
    {'Z': 57, 'name': "lanthanum", 'id': "La", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 58, 'name': "cerium", 'id': "Ce", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 59, 'name': "praseodymium", 'id': "Pr", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 60, 'name': "neodymium", 'id': "Nd", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 61, 'name': "promethium", 'id': "Pm", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "From decay"},
    {'Z': 62, 'name': "samarium", 'id': "Sm", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 63, 'name': "europium", 'id': "Eu", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 64, 'name': "gadolinium", 'id': "Gd", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 65, 'name': "terbium", 'id': "Tb", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 66, 'name': "dysprosium", 'id': "Dy", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 67, 'name': "holmium", 'id': "Ho", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 68, 'name': "erbium", 'id': "Er", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 69, 'name': "thulium", 'id': "Tm", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},
    {'Z': 70, 'name': "ytterbium", 'id': "Yb", 'phase': "Solid",
        'category': "Lanthanide", 'occurrence': "Primordial"},

], [
    (3, "**"),
    {'Z': 89, 'name': "actinium", 'id': "Ac", 'phase': "Solid",
        'category': "Actinide", 'occurrence': "From decay"},
    {'Z': 90, 'name': "thorium", 'id': "Th", 'phase': "Solid",
        'category': "Actinide", 'occurrence': "Primordial"},
    {'Z': 91, 'name': "protactinium", 'id': "Pa", 'phase': "Solid",
        'category': "Actinide", 'occurrence': "From decay"},
    {'Z': 92, 'name': "uranium", 'id': "U", 'phase': "Solid",
        'category': "Actinide", 'occurrence': "Primordial"},
    {'Z': 93, 'name': "neptunium", 'id': "Np", 'phase': "Solid",
        'category': "Actinide", 'occurrence': "From decay"},
    {'Z': 94, 'name': "plutonium", 'id': "Pu", 'phase': "Solid",
        'category': "Actinide", 'occurrence': "Primordial"},
    {'Z': 95, 'name': "americium", 'id': "Am", 'phase': "Solid",
        'category': "Actinide", 'occurrence': "From decay"},
    {'Z': 96, 'name': "curium", 'id': "Cm", 'phase': "Solid",
        'category': "Actinide", 'occurrence': "From decay"},
    {'Z': 97, 'name': "berkelium", 'id': "Bk", 'phase': "Solid",
        'category': "Actinide", 'occurrence': "From decay"},
    {'Z': 98, 'name': "californium", 'id': "Cf", 'phase': "Solid",
        'category': "Actinide", 'occurrence': "From decay"},
    {'Z': 99, 'name': "einsteinium", 'id': "Es", 'phase': "Solid",
        'category': "Actinide", 'occurrence': "Synthetic"},
    {'Z': 100, 'name': "fermium", 'id': "Fm", 'phase': "Unknown phase",
        'category': "Actinide", 'occurrence': "Synthetic"},
    {'Z': 101, 'name': "mendelevium", 'id': "Md", 'phase': "Unknown phase",
        'category': "Actinide", 'occurrence': "Synthetic"},
    {'Z': 102, 'name': "nobelium", 'id': "No", 'phase': "Unknown phase",
        'category': "Actinide", 'occurrence': "Synthetic"}
]]



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Mon Dec 08 23:57:48 2014

@author: Vidar Tonaas Fauske
"""


class ProcessCanceled(Exception):
    pass



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.

from hyperspyui.plugins.plugin import Plugin
import matplotlib as mpl
import matplotlib.animation as animation
import os
import sys

from qtpy import QtCore, QtWidgets
from qtpy.QtWidgets import QLineEdit, QCheckBox


def tr(text):
    return QtCore.QCoreApplication.translate("MovieSaver", text)


# =============================================================================
# Check that we have a valid writer
writer = mpl.rcParams['animation.writer']
writers = animation.writers
if writer in writers.avail:
    writer = writers[writer]()
else:
    import warnings
    warnings.warn("MovieWriter %s unavailable" % writer)

    try:
        writer = writers[writers.list()[0]]()
    except IndexError:
        raise ValueError("Cannot save animation: no writers are "
                         "available. Please install mencoder or "
                         "ffmpeg to save animations.")
del writer
# =============================================================================


class MovieSaver(Plugin):
    name = "Movie Saver"

    def create_actions(self):
        self.add_action(self.name + '.save', self.name, self.save,
                        icon="../images/video.svg",
                        tip="")

    def create_menu(self):
        self.add_menuitem('File', self.ui.actions[self.name + '.save'])

    def create_toolbars(self):
        self.add_toolbar_button(
            'File',
            self.ui.actions[
                self.name +
                '.save'])

    def save(self, wrapper=None, fps=None, fname=None, dpi=None):
        if wrapper is None:
            wrapper = self.ui.get_selected_wrapper()
        signal = wrapper.signal

        # TODO: Input: writer type, FPS, file, resolution/dpi [, bitrate ++]
        dlg = MovieArgsPrompt(self.ui)
        fname = os.path.join(os.path.dirname(self.ui.cur_dir), wrapper.name)
        fname += os.path.extsep + "mp4"
        dlg.edt_fname.setText(fname)
        dlg_w = self.ui.show_okcancel_dialog(tr("Save movie"), dlg)
        if dlg_w.result() == QtWidgets.QDialog.Accepted:
            # Setup writer:
            fps = dlg.num_fps.value()
            codec = dlg.edt_codec.text()
            if not codec:
                codec = None
            extra = dlg.edt_extra.text()
            if extra:
                extra = list(extra.split())
            else:
                extra = None
            if dlg.chk_verbose.isChecked():
                old_verbose = mpl.verbose.level
                mpl.verbose.level = 'debug'
                if extra:
                    extra.extend(['-v', 'debug'])
                else:
                    extra = ['-v', 'debug']
            metadata = signal.metadata.as_dictionary()
            writer = mpl.rcParams['animation.writer']
            writers = animation.writers
            if writer in writers.avail:
                writer = writers[writer](fps=fps, metadata=metadata,
                                         codec=codec, extra_args=extra)
            else:
                import warnings
                warnings.warn("MovieWriter %s unavailable" % writer)

                try:
                    writer = writers[writers.list()[0]](fps=fps,
                                                        metadata=metadata)
                except IndexError:
                    raise ValueError("Cannot save animation: no writers are "
                                     "available. Please install mencoder or "
                                     "ffmpeg to save animations.")

            fname = dlg.edt_fname.text()
            dpi = dlg.num_dpi.value()
            fig = signal._plot.signal_plot.figure

            # Set figure props:
            if not dlg.chk_colorbar.isChecked():
                cb_ax = signal._plot.signal_plot._colorbar.ax
                fig.delaxes(cb_ax)
            if not dlg.chk_axes.isChecked():
                signal._plot.signal_plot.ax.set_axis_off()

            try:
                with writer.saving(fig, fname, dpi):
                    for idx in signal.axes_manager:
                        QtWidgets.QApplication.processEvents()
                        writer.grab_frame()
            finally:
                # Reset figure props:
                if dlg.chk_verbose.isChecked():
                    mpl.verbose.level = old_verbose
                if not dlg.chk_colorbar.isChecked():
                    fig.add_axes(cb_ax)
                if not dlg.chk_axes.isChecked():
                    signal._plot.signal_plot.ax.set_axis_on()
                fig.canvas.draw()


class MovieArgsPrompt(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(MovieArgsPrompt, self).__init__(parent)
        self.create_controls()

    def create_controls(self):
        self.num_fps = QtWidgets.QDoubleSpinBox()
        self.num_fps.setValue(15.0)
        self.num_fps.setMinimum(0.001)

        self.edt_fname = QLineEdit()
        self.btn_browse = QtWidgets.QPushButton("...")

        self.num_dpi = QtWidgets.QSpinBox()
        self.num_dpi.setValue(72)
        self.num_dpi.setMinimum(1)
        self.num_dpi.setMaximum(10000)

        self.chk_axes = QCheckBox("Axes")
        self.chk_colorbar = QCheckBox("Colorbar")

        # codec = mpl.rcParams['animation.codec']
        codec = 'h264'
        self.edt_codec = QLineEdit(codec)
        self.edt_extra = QLineEdit("-preset veryslow -crf 0")

        # TODO: Use QCompleter or QComboBox for codecs
        # TODO: Use QCompleter for 'extra' history
        # TODO: Bitrate and/or quality slider

        self.chk_verbose = QCheckBox("Verbose")
        try:
            sys.stdout.fileno()
        except:
            self.chk_verbose.setEnabled(False)
            self.chk_verbose.setToolTip("Verbose output does not work with " +
                                        "internal console.")

        frm = QtWidgets.QFormLayout()
        frm.addRow(tr("FPS:"), self.num_fps)
        frm.addRow(tr("DPI:"), self.num_dpi)
        frm.addRow(tr("Codec:"), self.edt_codec)
        frm.addRow(tr("Extra args:"), self.edt_extra)
        frm.addRow(self.chk_axes, self.chk_colorbar)
        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(self.edt_fname)
        hbox.addWidget(self.btn_browse)
        frm.addRow(tr("File:"), hbox)
        frm.addRow("", self.chk_verbose)

        self.setLayout(frm)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtWidgets

from hyperspyui.widgets.signallist import SignalList


class PickXSignalsWidget(QtWidgets.QWidget):

    def __init__(self, signal_wrappers, x, parent=None, titles=None,
                 wrap_col=4):
        super(PickXSignalsWidget, self).__init__(parent)
        grid = QtWidgets.QGridLayout()
        self.pickers = []
        self._x = x
        self._sws = signal_wrappers

        if titles is None:
            titles = list(("",) * x)
        elif isinstance(titles, str):
            titles = list((titles,) * x)
        elif len(titles) < x:
            titles.extend(list(("",) * (x-len(titles))))

        for i in range(x):
            picker = SignalList(signal_wrappers, self, False)
            grid.addWidget(QtWidgets.QLabel(titles[i]), 2*(i // wrap_col),
                           i % wrap_col)
            grid.addWidget(picker, 1 + 2*(i // wrap_col), i % wrap_col)
            self.pickers.append(picker)
        self.setLayout(grid)

    def get_selected(self):
        signals = []
        for i in range(self._x):
            signals.append(self.pickers[i].get_selected())
        if self._x == 1:
            return signals[0]
        return signals

    def unbind(self):
        for i in range(self._x):
            self.pickers[i].unbind(self._sws)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Mar 01 15:18:30 2015

@author: Vidar Tonaas Fauske
"""

from hyperspyui.plugins.plugin import Plugin

from qtpy import QtCore
from qtpy.QtWidgets import QMessageBox

import hyperspy.utils.plot


def tr(text):
    return QtCore.QCoreApplication.translate("MirrorPlotPlugin", text)


class MirrorPlotPlugin(Plugin):
    name = "Mirror plot"

    def create_actions(self):
        self.add_action('mirror', "Mirror navigation", self.mirror_navi,
                        icon='mirror.svg',
                        selection_callback=self.ui.select_signal,
                        tip="Mirror navigation axes of selected signals")
        self.add_action('share_nav', "Share navigation", self.share_navi,
                        icon='intersection.svg',
                        selection_callback=self.ui.select_signal,
                        tip="Mirror navigation axes of selected signals")

    def create_menu(self):
        self.add_menuitem('Signal', self.ui.actions['mirror'])
        self.add_menuitem('Signal', self.ui.actions['share_nav'])

    def create_toolbars(self):
        self.add_toolbar_button("Signal", self.ui.actions['mirror'])
        self.add_toolbar_button("Signal", self.ui.actions['share_nav'])

    def share_navi(self, uisignals=None):
        self.mirror_navi(uisignals, shared_nav=True)

    def mirror_navi(self, uisignals=None, shared_nav=False):
        # Select signals
        if uisignals is None:
            uisignals = self.ui.get_selected_wrappers()
        if len(uisignals) < 2:
            mb = QMessageBox(QMessageBox.Information, tr("Select two or more"),
                             tr("You need to select two or more signals" +
                                " to mirror"), QMessageBox.Ok)
            mb.exec_()
            return

        signals = [s.signal for s in uisignals]

        # hyperspy closes, and then recreates figures when mirroring
        # the navigators. To keep UI from flickering, we suspend updates.
        # SignalWrapper also saves and then restores window geometry
        self.ui.setUpdatesEnabled(False)
        try:
            if shared_nav:
                navs = ["auto"]
                navs.extend([None] * (len(signals)-1))
                hyperspy.utils.plot.plot_signals(signals, sync=True,
                                                 navigator_list=navs)
            else:
                hyperspy.utils.plot.plot_signals(signals, sync=True)
        finally:
            self.ui.setUpdatesEnabled(True)    # Continue updating UI



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Tue Nov 04 13:37:08 2014

@author: Vidar Tonaas Fauske
"""

import os
import gc
import re
import numpy as np
import warnings
import traceback
import sys

from .mainwindowutillayer import MainWindowActionRecorder, tr

from . import uiprogressbar
uiprogressbar.takeover_progressbar()    # Enable hooks
from . import hooksignal
hooksignal.hook_signal()

from qtpy import QtCore, QtWidgets
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QMessageBox, QLabel

from hyperspyui.signalwrapper import SignalWrapper
from hyperspyui.bindinglist import BindingList
from hyperspyui.log import logger
from hyperspyui.widgets.dataviewwidget import DataViewWidget
from hyperspyui.widgets.editorwidget import EditorWidget
import hyperspyui.util
from hyperspyui.mdi_mpl_backend import FigureCanvas

import hyperspy.defaults_parser

from . import overrides
overrides.override_hyperspy()           # Enable hyperspy overrides


glob_escape = re.compile(r'([\[\]])')


class TrackEventFilter(QtCore.QObject):
    """Qt Event filter for tracking the mouse position in the application.
    """

    track = Signal(QtCore.QPoint)

    def eventFilter(self, receiver, event):
        if(event.type() == QtCore.QEvent.MouseMove):
            self.track.emit(event.globalPos())
        # Call Base Class Method to Continue Normal Event Processing
        return False


class MainWindowHyperspy(MainWindowActionRecorder):

    """
    Fifth layer in the application stack. Should integrate hyperspy basics,
    such as UI wrappings for hyperspy classes (Signal and Model), file I/O,
    etc.
    """

    def __init__(self, parent=None):
        # Setup signals list. This is a BindingList, and all components of the
        # code that needs to keep track of the signals loaded bind into this.
        self.signals = BindingList()
        self.signals.add_custom(self._sweeper, None, None, None, self._sweeper,
                                None)
        self.hspy_signals = []

        def rem(x):
            # Needed since hyperspy has special equality operator...
            for i, s in enumerate(self.hspy_signals):
                if s is x.signal:
                    self.hspy_signals.pop(i)
        self.signals.add_custom(
            'hspy_signals',
            lambda x: self.hspy_signals.append(x.signal),
            lambda x, y: self.hspy_signals.insert(x, y.signal),
            None,
            rem,
            lambda x: self.hspy_signals.pop(x))
        self.lut_signalwrapper = dict()

        def lut_add(sw):
            self.lut_signalwrapper[sw.signal] = sw
        lut = self.lut_signalwrapper
        self.signals.add_custom('lut', lut_add, None, None,
                                lambda sw: lut.pop(sw.signal), None)

        # Setup variables
        self.progressbars = {}
        self.prev_mdi = None
        self._plotting_signal = None

        # Call super init, which creates main controls etc.
        super(MainWindowHyperspy, self).__init__(parent)

        self.create_statusbar()

        # Enable drag and drop
        self.setAcceptDrops(True)

        # Connect UIProgressBar for graphical hyperspy progress
        s = uiprogressbar.signaler
        s.created[int, int, str].connect(self.on_progressbar_wanted)
        s.progress[int, int, str].connect(self.on_progressbar_update)
        s.finished[int].connect(self.on_progressbar_finished)
        self.cancel_progressbar.connect(s.on_cancel)

        # Connect to Signal.plot events
        hooksignal.connect_plotting(self.on_signal_plotting)
        hooksignal.connect_plotted(self.on_signal_plotted)

        # Finish off hyperspy customization of layer 1
        self.setWindowTitle("HyperSpy")

    def _sweeper(self, removed):
        """
        Trigger a GC one second after calling this.
        """
        del removed
        QtCore.QTimer.singleShot(1000, gc.collect)

    def create_widgetbar(self):
        # Add DataViewWidget to widget bar:
        self.tree = DataViewWidget(self, self)
        self.tree.setWindowTitle(tr("Data View"))
        # Sync tree with signals list:
        self.signals.add_custom(self.tree, self.tree.add_signal, None,
                                None, self.tree.remove, None)
        self.main_frame.subWindowActivated.connect(
            self.tree.on_mdiwin_activated)
        self.add_widget(self.tree)

        # Put other widgets at end (plugin widgets)
        super(MainWindowHyperspy, self).create_widgetbar()

    def create_menu(self):
        # Super creates Windows menu
        super(MainWindowHyperspy, self).create_menu()

        # Add custom action to signals' BindingList, so appropriate menu items
        # are removed if a signal is removed from the list
        def rem_s(value):
            for f in value.figures:
                self.windowmenu.removeAction(f.activateAction())
        self.signals.add_custom(self.windowmenu, None, None, None,
                                rem_s, lambda i: rem_s(self.signals[i]))

    def create_statusbar(self):
        """
        Creates extra status bar controls, e.g. coordinate tracking.
        """
        sb = self.statusBar()
        self.nav_coords_label = QLabel("Navigation: ()")
        sb.addPermanentWidget(self.nav_coords_label)
        self.mouse_coords_label = QLabel("Mouse: (,) px; (,)")
        sb.addPermanentWidget(self.mouse_coords_label)

        # To be able to update coordinates, we need to track the mouse
        # position with an event filter
        self.main_frame.subWindowActivated.connect(
            self._connect_figure_2_statusbar)
        app = QtWidgets.QApplication.instance()
        self.tracker = TrackEventFilter()
        self.tracker.track.connect(self._on_track)
        app.installEventFilter(self.tracker)

    def _connect_figure_2_statusbar(self, mdi_window):
        """
        When a figure is activated, this callback sets the navigation status
        bar, and connects _on_active_navigate to the signal's AxesManager.
        """
        if mdi_window is self.prev_mdi:
            return
        s = hyperspyui.util.win2sig(mdi_window)
        if self.prev_mdi is not None:
            ps = hyperspyui.util.win2sig(self.prev_mdi, self.signals)
        else:
            ps = None
        if s is not ps:
            # If previous signal present, try to disconnect
            if ps and ps.signal and ps.signal.axes_manager:
                try:
                    ps.signal.axes_manager.events.indices_changed.disconnect(
                        self._on_active_navigate)
                except ValueError:
                    pass
            if s and s.signal and s.signal.axes_manager:
                try:
                    s.signal.axes_manager.events.indices_changed.connect(
                        self._on_active_navigate, {'obj': 'axes_manager'})
                except ValueError:
                    pass
                self._on_active_navigate(s.signal.axes_manager)
            self.prev_mdi = mdi_window

    def _on_active_navigate(self, axes_manager):
        """
        Callback triggered when the active signal navigates. Updates the
        status bar with the navigation indices.
        """
        ind = axes_manager.indices
        self.set_navigator_coords_status(ind)

    def _on_track(self, gpos):
        """
        Tracks the mouse position for the entire application, and if the mouse
        is over a figure axes it updates the status bar mouse coordinates.
        """

        # Find which window the mouse is above
        pos = self.mapFromGlobal(gpos)
        canvas = self.childAt(pos)
        # We only care about FigureCanvases
        if isinstance(canvas, FigureCanvas):
            fig = canvas.figure
            s, p = hyperspyui.util.fig2sig(fig, self.signals)
            # Currently we only know how to deal with standard plots
            if p is None:
                return
        else:
            return
        if p.ax is None:
            return
        # Map position to canvas frame of reference
        cpos = canvas.mapFromGlobal(gpos)
        # Mapping copied from MPL backend code:
        cpos = np.array([(cpos.x(), canvas.figure.bbox.height - cpos.y())])
        # Check that we are within plot axes
        (xa, ya), = p.ax.transAxes.inverted().transform(cpos)
        if not (0 <= xa <= 1 and 0 <= ya <= 1):
            return
        # Find coordinate values:
        (xd, yd), = p.ax.transData.inverted().transform(cpos)

        def v2i(a, v):
            if v < a.low_value:
                return a.low_index
            elif v > a.high_value:
                return a.high_index
            return a.value2index(v)

        if hasattr(p, 'axis'):                              # SpectrumFigure
            if p is s.signal._plot.navigator_plot:
                axis = p.axes_manager.navigation_axes[0]
            elif p is s.signal._plot.signal_plot:
                axis = p.axes_manager.signal_axes[0]
            else:
                return
            vals = (xd,)
            ind = (v2i(axis, xd),)
            units = [axis.units]
            intensity = yd
        elif hasattr(p, 'xaxis') and hasattr(p, 'yaxis'):   # ImagePlot
            vals = (xd, yd)
            ind = (v2i(p.xaxis, xd),
                   v2i(p.yaxis, yd))
            units = [p.xaxis.units, p.yaxis.units]
            intensity = p.ax.images[0].get_array()[ind[1], ind[0]]

        # Remove <undefined> units
        for i in range(len(units)):
            if str(units[i]) == "<undefined>":
                units[i] = ""
        units = tuple(units)

        # Finally, display coordinates
        self.set_mouse_coords_status(ind, vals, units, intensity)

    def set_navigator_coords_status(self, coords):
        """
        Displays 'coords' as the navigator coordinates.
        """
        self.nav_coords_label.setText("Navigation: " + str(coords))

    def set_mouse_coords_status(self, indices, values, units, intensity=None):
        """
        Display mouse coordinates both in indices and data space values.

        'units' must be the same size as 'values'
        """
        vu = tuple(["%.3g %s" % (v, u) for v, u in zip(values, units)])
        vu = "(%s)" % ", ".join(vu)
        text = "Mouse: " + str(indices) + " px; " + vu
        if intensity is not None:
            text += "; Intensity: "
            if isinstance(intensity, np.ndarray):
                text += str(intensity)
            else:
                text += "%.3g" % intensity
        self.mouse_coords_label.setText(text)

    def add_model(self, signal=None, *args, **kwargs):
        """
        Add a default model for the given/selected signal. Returns the
        newly created ModelWrapper.
        """
        if signal is None:
            signal = self.get_selected_wrapper()
        elif not isinstance(signal, SignalWrapper):
            signal = [s for s in self.signals if s.signal is signal]
            signal = signal[0]
        mw = signal.make_model(*args, **kwargs)
        return mw

    def make_component(self, comp_type):
        m = self.get_selected_model_wrapper()
        if m is None:
            sw = self.get_selected_wrapper()
            if sw is None:
                return
            m = self.add_model(signal=sw)
        m.add_component(comp_type)

    def edit_hspy_settings(self):
        hyperspy.api.preferences.gui()

    # -------- Signal plotting callbacks -------
    def on_signal_plotting(self, signal, *args, **kwargs):
        # Check if we have a wrapper, if not we make one:
        if signal in self.lut_signalwrapper:
            # Replotting, make sure we keep it when closing
            sw = self.lut_signalwrapper[signal]
            sw.keep_on_close = True
        else:
            # New signal, make wrapper and add to list
            sw = SignalWrapper(signal, self)
            self.signals.append(sw)
        self._plotting_signal = sw

    def on_signal_plotted(self, signal, *args, **kwargs):
        sw = self.lut_signalwrapper[signal]
        sw.update_figures()
        if sw.keep_on_close:
            sw.keep_on_close = False
        self._plotting_signal = None
        self.main_frame.subWindowActivated.emit(
            self.main_frame.activeSubWindow())

    # -------- Selection management -------

    def get_selected_wrapper(self, error_on_multiple=False):
        signals = self.get_selected_wrappers()
        if signals is None or len(signals) < 1:
            return None
        elif len(signals) == 1:
            return signals[0]
        else:
            if error_on_multiple:
                mb = QMessageBox(QMessageBox.Information,
                                 tr("Select one signal only"),
                                 tr("You can only select one signal at the " +
                                     "time for this function. Currently, " +
                                     "several are selected"),
                                 QMessageBox.Ok)
                mb.exec_()
                raise RuntimeError()
            w = self.main_frame.activeSubWindow()
            s = [hyperspyui.util.win2sig(w, self.signals)]
            if s in signals:
                return s
            else:
                return signals[0]

    def get_selected_wrappers(self):
        s = self.tree.get_selected_wrappers()
        if len(s) < 1:
            w = self.main_frame.activeSubWindow()
            s = [hyperspyui.util.win2sig(w, self.signals)]
        return s

    def get_selected_signals(self):
        try:
            return [s.signal for s in self.get_selected_wrappers()]
        except AttributeError:
            logger.info("No signal available.")

    def get_selected_signal(self):
        sw = self.get_selected_wrapper()
        if sw is None:
            logger.info("No signal available.")
        else:
            return sw.signal

    def get_selected_model_wrapper(self):
        """
        Returns the selected model
        """
        return self.tree.get_selected_model()

    def get_selected_model(self):
        """
        Returns the selected model
        """
        mw = self.get_selected_model_wrapper()
        return None if mw is None else mw.model

    def get_selected_component(self):
        """
        Returns the selected component
        """
        return self.tree.get_selected_component()

    def get_selected_plot(self):
        """
        Returns the selected signal; a string specifying whether the active
        window is "navtigation" plot, "signal" plot or "other"; and finally the
        active window.
        """
        s = self.get_selected_wrapper()
        if s is None:
            logger.info("No plot available.")
            return
        w = self.main_frame.activeSubWindow()
        if w is s.navigator_plot:
            selected = "navigation"
        elif w is s.signal_plot:
            selected = "signal"
        else:
            selected = "other"
        return s, selected, w

    def select_signal(self, win, action):
        """Signal selection callback for actions that are only valid for
        selected Signals.
        """
        s = hyperspyui.util.win2sig(win, self.signals, self._plotting_signal)
        if s is None:
            action.setEnabled(False)
        else:
            action.setEnabled(True)

    def select_model(self, win, action):
        """Model selection callback for actions that are only valid for
        selected Models.
        """
        m = self.get_selected_model()
        action.setEnabled(m is not None)

    # --------- File I/O ----------

    @staticmethod
    def get_accepted_extensions():
        from hyperspy.io_plugins import io_plugins
        extensions = set([extensions.lower() for plugin in io_plugins
                          for extensions in plugin.file_extensions])
        return extensions

    def load_stack(self, filenames=None, stack_axis=None):
        if filenames is None:
            extensions = self.get_accepted_extensions()
            type_choices = ';;'.join(["*." + e for e in extensions])
            type_choices = ';;'.join(("Python code (*.py)", type_choices))
            type_choices = ';;'.join(("All types (*.*)", type_choices))
            filenames = self.prompt_files(type_choices)
            if not filenames:
                return False
            self.cur_dir = filenames[0]
        for i, f in enumerate(filenames):
            filenames[i] = glob_escape.sub(r'[\1]', f)    # glob escapes

        sig = hyperspy.io.load(filenames, stack=True, stack_axis=stack_axis)
        if isinstance(sig, list):
            for s in sig:
                s.plot()
        else:
            sig.plot()
        self.record_code('ui.load_stack({0}, {1})'.format(filenames,
                                                          stack_axis))

    def load(self, filenames=None):
        """
        Load 'filenames', or if 'filenames' is None, open a dialog to let the
        user interactively browse for files. It then load these files using
        hyperspy.io.load and wraps them and adds them to self.signals.
        """

        import hyperspy.io
        if filenames is None:
            extensions = self.get_accepted_extensions()
            type_choices = ';;'.join(["*." + e for e in extensions])
            type_choices = ';;'.join(("Python code (*.py)", type_choices))
            type_choices = ';;'.join(("All types (*.*)", type_choices))
            filenames = self.prompt_files(type_choices)
            if not filenames:
                return False
            self.cur_dir = filenames[0]

        files_loaded = []
        for filename in filenames:
            self.set_status("Loading \"" + filename + "\"...")
            ext = os.path.splitext(filename)[1]
            if ext == '.py':
                e = EditorWidget(self, self, filename)
                self.editors.append(e)
                e.show()
                files_loaded.append(filename)
                continue
            self.setUpdatesEnabled(False)   # Prevent flickering during load
            try:
                escaped = glob_escape.sub(r'[\1]', filename)    # glob escapes
                sig = hyperspy.io.load(escaped)
                if isinstance(sig, list):
                    for s in sig:
                        s.plot()
                else:
                    sig.plot()
                files_loaded.append(filename)
            except (IOError, ValueError) as e:
                self.set_status("Failed to load \"" + filename + "\"")
                exc_type, exc_value, exc_traceback = sys.exc_info()
                tb = traceback.extract_tb(exc_traceback)[-1]
                warnings.warn_explicit(
                    ("Failed to load file '%s'. Internal exception:\n %s: %s"
                     % (filename, exc_type.__name__, str(e))),
                    RuntimeWarning, tb[0], tb[1])
            finally:
                self.setUpdatesEnabled(True)    # Always resume updates!

        if len(files_loaded) == 1:
            self.set_status("Loaded \"" + files_loaded[0] + "\"")
        elif len(files_loaded) > 1:
            self.set_status("Loaded %d files" % len(files_loaded))
        self.record_code('ui.load({0})'.format(files_loaded))

        return files_loaded

    def save(self, signals=None, filenames=None):
        logger.debug("entering save(), with args: %s, %s",
                     str(signals), str(filenames))
        if signals is None:
            signals = self.get_selected_wrappers()
            logger.debug("No signals passed, saving selection: %s",
                         str(signals))

        extensions = self.get_accepted_extensions()
        type_choices = ';;'.join(["*." + e for e in extensions])
        type_choices = ';;'.join(("All types (*.*)", type_choices))
        logger.debug("Save type choices: %s", type_choices)

        i = 0
        overwrite = None
        for s in signals:
            # Match signal to filename. If filenames has not been specified,
            # or there are no valid filename for curren signal index i, we
            # have to prompt the user.
            if filenames is None or len(
                    filenames) <= i or filenames[i] is None:
                path_suggestion = self.get_signal_filepath_suggestion(s)
                logger.debug("No filenames passed. Auto-suggestion: %s",
                             path_suggestion)
                filename = self.prompt_files(type_choices, path_suggestion,
                                             False)
                # Dialog should have prompted about overwrite
                overwrite = True
                if not filename:
                    logger.info("Not saving signal %s", str(s))
                    continue
            else:
                filename = filenames[i]
                overwrite = None    # We need to confirm overwrites
            i += 1
            s.signal.save(filename, overwrite)

    def get_signal_filepath_suggestion(self, signal, default_ext=None):
        if default_ext is None:
            try:
                # hspy < 1.3
                default_ext = hyperspy.defaults_parser.preferences.General.\
                             default_file_format
            except AttributeError:
                # hspy >= 1.3
                default_ext = 'hspy'
        # Get initial suggestion for save dialog.  Use
        # original_filename metadata if present, or self.cur_dir if not
        if signal.signal.metadata.has_item('General.original_filename'):
            f = signal.signal.metadata.General.original_filename
        else:
            f = self.cur_dir

        # Analyze suggested filename
        base, tail = os.path.split(f)
        fn, ext = os.path.splitext(tail)

        # If no directory in filename, use self.cur_dir's dirname
        if base is None or base == "":
            base = os.path.dirname(self.cur_dir)
        # If extension is not valid, use the defualt
        extensions = self.get_accepted_extensions()
        if ext not in extensions:
            ext = default_ext
        # Filename itself is signal's name
        fn = signal.name
        if os.name == 'nt':
            fn = fn.replace("<", "[").replace(">", "]")
            fn = re.sub(r"[:\"|\?\*]", '', fn)
        # Build suggestion and return
        path_suggestion = os.path.sep.join((base, fn))
        path_suggestion = os.path.extsep.join((path_suggestion, ext))
        return path_suggestion

    # ---------- Drag and drop overloads ----------

    def dragEnterEvent(self, event):
        # Check file name extensions to see if we should accept
        extensions = set(self.get_accepted_extensions().union(('py',)))
        mimeData = event.mimeData()
        if mimeData.hasUrls():
            pathList = [url.toLocalFile() for url in mimeData.urls()]
            data_ext = set([os.path.splitext(p)[1][1:] for p in pathList])
            # Accept as long as we can read some of the files being dropped
            if 0 < len(data_ext.intersection(extensions)):
                event.acceptProposedAction()

#    def dragMoveEvent(event):
#        pass
#
#    def dragLeaveEvent(event):
#        pass

    def dropEvent(self, event):
        # Something has been dropped. Try to load all file urls
        mimeData = event.mimeData()
        if mimeData.hasUrls():
            pathList = [url.toLocalFile() for url in mimeData.urls()]
            if self.load(pathList):
                event.acceptProposedAction()

    # --------- Hyperspy progress bars ----------

    cancel_progressbar = Signal(int)

    def on_progressbar_wanted(self, pid, maxval, label):
        logger.debug("Progressbar wanted. ID: %d", pid)
        if pid in self.progressbars:
            progressbar = self.progressbars[pid]
            progressbar.setValue(0)
        else:
            progressbar = QtWidgets.QProgressDialog(self)
            progressbar.setMinimumDuration(2000)
            progressbar.setMinimum(0)
        progressbar.setMaximum(maxval)
        progressbar.setWindowTitle("Processing")
        progressbar.setLabelText(label)

        if pid not in self.progressbars:
            def cancel():
                # If pid not in collection, it is finished, and cancel
                # triggered when dialog closed.
                if pid in self.progressbars:
                    logger.debug("Progressbar canceled. ID: %d", pid)
                    self.cancel_progressbar.emit(pid)

            progressbar.canceled.connect(cancel)
            progressbar.setWindowModality(QtCore.Qt.WindowModal)

            self.progressbars[pid] = progressbar

    def on_progressbar_update(self, pid, value, txt=None):
        if pid not in self.progressbars:
            return
        self.progressbars[pid].setValue(value)
        if txt is not None:
            self.progressbars[pid].setLabelText(txt)

    def on_progressbar_finished(self, pid):
        logger.debug("Progressbar finished. ID: %d", pid)
        progressbar = self.progressbars.pop(pid)
        progressbar.close()

    # --------- End hyperspy progress bars ----------

    # --------- Console functions ----------

    def on_console_executing(self, source):
        super(MainWindowHyperspy, self).on_console_executing(source)
        #self.setUpdatesEnabled(False)
        for s in self.signals:
            s.keep_on_close = True

    def on_console_executed(self, response):
        super(MainWindowHyperspy, self).on_console_executed(response)
        for s in self.signals:
            s.update_figures()
            s.keep_on_close = False
        #self.setUpdatesEnabled(True)

    def _get_console_exec(self):
        ex = super(MainWindowHyperspy, self)._get_console_exec()
        ex += '\nimport hyperspy.api as hs'
        ex += '\nimport numpy as np'
        return ex

    def _get_console_exports(self):
        push = super(MainWindowHyperspy, self)._get_console_exports()
        push['siglist'] = self.hspy_signals
        return push

    def _get_console_config(self):
        # ===== THIS ======
        try:
            from traitlets.config.loader import PyFileConfigLoader
        except ImportError:
            from IPython.config.loader import PyFileConfigLoader
        ipcp = os.path.sep.join((os.path.dirname(__file__), "ipython_profile",
                                 "ipython_embedded_config.py"))
        c = PyFileConfigLoader(ipcp).load_config()
        # ===== OR THIS =====
#        import hyperspy.Release
#        try:
#            from traitlets.config import Config
#        except ImportError:
#            from IPython.config import Config
#        c = Config()
#        c.FrontendWidget.banner = hyperspy.Release.info
        # ===== END =====
        return c



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Wed Oct 29 16:49:48 2014

@author: Vidar Tonaas Fauske
"""


from qtpy import QtCore, QtWidgets

from .extendedqwidgets import FigureWidget, ExDoubleSlider, ExClickLabel
from hyperspy.misc.rgb_tools import rgbx2regular_array
from hyperspyui.util import win2fig, fig2image_plot, block_signals

import numpy as np
from matplotlib.colors import Normalize, SymLogNorm


def tr(text):
    return QtCore.QCoreApplication.translate("ContrastWidget", text)


def _auto_contrast(plot):
    return [plot._vmin_user is None, plot._vmax_user is None]


def _set_auto_contrast(plot, auto):
    if auto:
        plot._vmin_user = None
        plot._vmax_user = None
    else:
        plot._vmin_user = plot.vmin
        plot._vmax_user = plot.vmax


class ContrastWidget(FigureWidget):
    LevelLabel = tr("Level")
    WindowLabel = tr("Window")

    def __init__(self, main_window, parent, figure=None):
        super(ContrastWidget, self).__init__(main_window, parent)
        self.setWindowTitle(tr("Contrast control"))
        self.create_controls()

        self._on_figure_change(figure)

    def _on_figure_change(self, figure):
        super(ContrastWidget, self)._on_figure_change(figure)
        if isinstance(figure, QtWidgets.QMdiSubWindow):
            figure = win2fig(figure)

        signals = self.parent().signals
        p = fig2image_plot(figure, signals)
        self._cur_plot = p
        self.update_controls_from_fig()

    def _plot_minmax(self):
        p = self._cur_plot
        if p is None:
            return 0.0, 0.0
        data = rgbx2regular_array(p.data_function().ravel())
        return np.nanmin(data), np.nanmax(data)

    def update_controls_from_fig(self):
        p = self._cur_plot
        if p is None:
            self.sl_level.setValue(0.0)
            self.sl_window.setValue(0.0)
            self.chk_log.setChecked(False)
            self.disable()
        else:
            self.enable()
            imin, imax = self._plot_minmax()

            with block_signals(self.sl_level):
                self.sl_level.setRange(imin, imax)
                self.sl_level.setValue(p.vmin)
                self.lbl_level.setText(self.LevelLabel + ": %.2G" % p.vmin)

            with block_signals(self.sl_window):
                self.sl_window.setRange(0, imax - imin)
                window = p.vmax - p.vmin
                self.sl_window.setValue(window)
                self.lbl_window.setText(self.WindowLabel + ": %.2G" % window)

            with block_signals(self.chk_auto):
                auto = _auto_contrast(p)
                state = QtCore.Qt.Checked if all(auto) else (
                    QtCore.Qt.PartiallyChecked if any(auto) else
                    QtCore.Qt.Unchecked)
                self.chk_auto.setCheckState(state)

            if p.ax.images:
                with block_signals(self.chk_log):
                    self.chk_log.setEnabled(True)
                    norm = isinstance(p.ax.images[-1].norm, SymLogNorm)
                    self.chk_log.setChecked(norm)
            else:
                self.chk_log.setEnabled(False)

    def level_changed(self, value):
        self.lbl_level.setText(self.LevelLabel + ": %.2G" % value)
        p = self._cur_plot
        if p is not None:
            p.vmin = self.sl_level.value()
            p.vmax = self.sl_level.value() + self.sl_window.value()
            if self.chk_auto.isChecked():
                self.chk_auto.setChecked(False)
            p.update()

    def window_changed(self, value):
        self.lbl_window.setText(self.WindowLabel + ": %.2G" % value)
        p = self._cur_plot
        if p is not None:
            p.vmax = self.sl_level.value() + self.sl_window.value()
            if self.chk_auto.isChecked():
                self.chk_auto.setChecked(False)
            p.update()

    def log_changed(self, value):
        p = self._cur_plot
        if p is not None and p.ax.images:
            old = p.ax.images[0].norm
            kw = dict(vmin=old.vmin, vmax=old.vmax, clip=old.clip)
            if value:
                n = SymLogNorm(1e-9, **kw)
            else:
                n = Normalize(**kw)
            p.ax.images[-1].norm = n
            p.update()

    def enable(self, enabled=True):
        self.lbl_level.setEnabled(enabled)
        self.lbl_window.setEnabled(enabled)
        self.sl_level.setEnabled(enabled)
        self.sl_window.setEnabled(enabled)
        self.chk_auto.setEnabled(enabled)
        self.chk_log.setEnabled(enabled)

    def disable(self):
        self.enable(False)

    def auto(self, state):
        p = self._cur_plot
        if p is not None:
            _set_auto_contrast(p, state == QtCore.Qt.Checked)
            p.update()
            self.update_controls_from_fig()

    def reset_level(self):
        imin, imax = self._plot_minmax()
        self.sl_level.setValue(imin)   # Will trigger events

    def reset_window(self):
        imin, imax = self._plot_minmax()
        self.sl_window.setValue(imax - imin)   # Will trigger events

    def create_controls(self):
        self.sl_level = ExDoubleSlider(self, QtCore.Qt.Horizontal)
        self.lbl_level = ExClickLabel(self.LevelLabel + ": 0.0")
        self.sl_window = ExDoubleSlider(self, QtCore.Qt.Horizontal)
        self.lbl_window = ExClickLabel(self.WindowLabel + ": 0.0")

        for sl in [self.sl_level, self.sl_window]:
            sl.setRange(0.0, 1.0)
            sl.setValue(0.0)

        self.chk_auto = QtWidgets.QCheckBox(tr("Auto"), self)
        self.chk_log = QtWidgets.QCheckBox(tr("Log"), self)

        # Test level/window working?
        self.lbl_level.clicked.connect(self.reset_level)
        self.lbl_window.clicked.connect(self.reset_window)
        self.sl_level.double_valueChanged.connect(self.level_changed)
        self.sl_window.double_valueChanged.connect(self.window_changed)
        self.chk_auto.stateChanged[int].connect(self.auto)
        self.chk_log.toggled[bool].connect(self.log_changed)

        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(self.chk_auto)
        hbox.addWidget(self.chk_log)
        vbox = QtWidgets.QVBoxLayout()
        for w in [self.sl_level, self.lbl_level,
                  self.sl_window, self.lbl_window]:
            vbox.addWidget(w)
        vbox.addLayout(hbox)

        wrap = QtWidgets.QWidget()
        wrap.setLayout(vbox)
        height = vbox.sizeHint().height()
        wrap.setFixedHeight(height)
        self.setWidget(wrap)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Mon Oct 27 18:47:05 2014

@author: Vidar Tonaas Fauske
"""


import hyperspy.components1d
from hyperspy.misc.utils import slugify
from functools import partial
from qtpy import QtGui, QtCore, QtSvg, QtWidgets
import os
from contextlib import contextmanager


def tr(text):
    return QtCore.QCoreApplication.translate("MainWindow", text)


@contextmanager
def dummy_context_manager(*args, **kwargs):
    yield


@contextmanager
def block_signals(target):
    old = target.blockSignals(True)
    try:
        yield
    finally:
        target.blockSignals(old)


def lstrip(string, prefix):
    if string is not None:
        if string.startswith(prefix):
            return string[len(prefix):]


def debug_trace():
    '''Set a tracepoint in the Python debugger that works with Qt'''
    from qtpy.QtCore import pyqtRemoveInputHook
    from pdb import set_trace
    pyqtRemoveInputHook()
    set_trace()


def fig2win(fig, windows):
    # Each figure has FigureCanvas as widget, canvas has figure property
    if fig is None:
        return None
    matches = [w for w in windows if w.widget().figure == fig]
    if len(matches) >= 1:
        return matches[0]
    else:
        return None


def fig2image_plot(fig, signals):
    from hyperspy.drawing.mpl_he import MPL_HyperExplorer
    from hyperspy.drawing.image import ImagePlot
    for s in signals:
        p = s.signal._plot
        if isinstance(p, MPL_HyperExplorer):
            if isinstance(p.signal_plot, ImagePlot):
                if p.signal_plot.figure is fig:
                    return p.signal_plot
            if isinstance(p.navigator_plot, ImagePlot):
                if p.navigator_plot.figure is fig:
                    return p.navigator_plot
    return None


def fig2sig(fig, signals):
    from hyperspy.drawing.mpl_he import MPL_HyperExplorer
    for s in signals:
        p = s.signal._plot
        if isinstance(p, MPL_HyperExplorer):
            if p.signal_plot and p.signal_plot.figure is fig:
                return s, p.signal_plot
            elif p.navigator_plot and p.navigator_plot.figure is fig:
                return s, p.navigator_plot
    return None, None


def win2fig(window):
    # Each figure has FigureCanvas as widget, canvas has figure property
    return window.widget().figure


def win2sig(window, signals=None, plotting_signal=None):
    r = None
    if window is not None:
        if signals is None:
            # Returns None if no such property
            r = window.property('hyperspyUI.SignalWrapper')
        else:
            for s in signals:
                if window in (s.navigator_plot, s.signal_plot):
                    return s
    if r is None and plotting_signal is not None:
        return plotting_signal
    return r


class SignalTypeFilter(object):

    def __init__(self, signal_type, ui, space=None):
        self.signal_type = signal_type
        self.ui = ui
        self.space = space

    def __call__(self, win, action):
        sig = win2sig(win, self.ui.signals, self.ui._plotting_signal)
        valid = sig is not None and isinstance(sig.signal, self.signal_type)
        if valid and self.space:
            # Check that we have right figure
            if not ((self.space == "navigation" and win is sig.navigator_plot)
                    or
                    (self.space == "signal" and win is sig.signal_plot)):
                valid = False
        action.setEnabled(valid)


def dict_rlu(dictionary, value):
    """
    Reverse dictionary lookup.
    """
    for k, v in dictionary.items():
        if v == value or v is value:
            return k
    raise KeyError()


def crosshair_cursor():
    if os.name == 'nt':
        # On windows, cursors support "inversion" mode, where they invert the
        # underlying color. This is achived with two bitmap.
        # Final cursor has mask all 0
        # When main bitmap is 0, this means transparent, when 1, inversion.
        bm = QtGui.QBitmap(16, 16)
        ma = QtGui.QBitmap(16, 16)
        bm.clear()
        ma.clear()
        # Paint a crosshair on the main bitmap with color1.
        pbm = QtGui.QPainter(bm)
        pbm.setPen(QtCore.Qt.color1)
        pbm.drawLine(8, 0, 8, 15)
        pbm.drawLine(0, 8, 15, 8)
        pbm.setPen(QtCore.Qt.color0)
        pbm.drawPoint(8, 8)
        pbm.end()
        return QtGui.QCursor(bm, ma, 8, 8)
    else:
        fn = os.path.dirname(__file__) + '/images/picker.svg'
        return load_cursor(fn, 8, 8)


def load_cursor(filename, hotX=-1, hotY=-1):
    renderer = QtSvg.QSvgRenderer(filename)
    pm = QtGui.QPixmap(16, 16)
    pm.fill(QtCore.Qt.transparent)
    painter = QtGui.QPainter(pm)
    renderer.render(painter)
    del painter, renderer
    return QtGui.QCursor(pm, hotX, hotY)


def create_add_component_actions(parent, callback, prefix="", postfix=""):
    actions = {}
    compnames = ['Arctan', 'Bleasdale', 'DoubleOffset', 'DoublePowerLaw',
                 'Erf', 'Exponential', 'Gaussian', 'GaussianHF', 'Logistic',
                 'Lorentzian', 'Offset', 'PowerLaw', 'SEE', 'RC', 'Vignetting',
                 'Voigt', 'Polynomial', 'PESCoreLineShape', 'Expression',
                 'VolumePlasmonDrude']
    for name in compnames:
        try:
            t = getattr(hyperspy.components1d, name)
        except AttributeError:
            continue
        ac_name = 'add_component_' + name
        f = partial(callback, t)
        ac = QtWidgets.QAction(prefix + name + postfix, parent)
        ac.setStatusTip(tr("Add a component of type ") + name)
        ac.triggered.connect(f)
        actions[ac_name] = ac
    return actions


class AttributeDict(dict):

    """A dict subclass that exposes its items as attributes.

    """

    def __init__(self, obj={}):
        super(AttributeDict, self).__init__(obj)

    def __dir__(self):
        return [slugify(k, True) for k in self]

    def __repr__(self):
        return "%s(%s)" % (
            type(self).__name__, super(Namespace, self).__repr__())

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            for k in self:
                if name == slugify(k, True):
                    return self[k]

    def __setattr__(self, name, value):
        if name in self:
            self[name] = value
        else:
            for k in self:
                if name == slugify(k, True):
                    self[k] = value
            else:
                self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            for k in self:
                if name == slugify(k, True):
                    del self[k]

    # ------------------------
    # "copy constructors"

    @classmethod
    def from_object(cls, obj, names=None):
        if names is None:
            names = dir(obj)
        ns = {name: getattr(obj, name) for name in names}
        return cls(ns)

    @classmethod
    def from_mapping(cls, ns, names=None):
        if names:
            ns = {name: ns[name] for name in names}
        return cls(ns)

    @classmethod
    def from_sequence(cls, seq, names=None):
        if names:
            seq = {name: val for name, val in seq if name in names}
        return cls(seq)

    # ------------------------
    # static methods

    @staticmethod
    def hasattr(ns, name):
        try:
            object.__getattribute__(ns, name)
        except AttributeError:
            return False
        return True

    @staticmethod
    def getattr(ns, name):
        return object.__getattribute__(ns, name)

    @staticmethod
    def setattr(ns, name, value):
        return object.__setattr__(ns, name, value)

    @staticmethod
    def delattr(ns, name):
        return object.__delattr__(ns, name)


class Namespace(AttributeDict):

    """A dict subclass that exposes its items as attributes.

    Warning: Namespace instances do not have direct access to the
    dict methods.

    """

    def __init__(self, obj={}):
        super(Namespace, self).__init__(obj)

    def __repr__(self):
        return "%s(%s)" % (
            type(self).__name__, dict.__repr__())

    def __getattribute__(self, name):
        try:
            return self[name]
        except KeyError:
            msg = tr("'%s' object has no attribute '%s'")
            raise AttributeError(msg % (type(self).__name__, name))



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sat Feb 28 17:52:42 2015

@author: Vidar Tonaas Fauske
"""

import hyperspy.signal
import hyperspy.events
orig_signal = hyperspy.signal.BaseSignal


class HookedSignal(orig_signal):

    def plot(self, *args, **kwargs):
        _on_plotting(self)
        r = super(HookedSignal, self).plot(*args, **kwargs)
        _on_plotted(self)
        return r


def hook_signal():
    """
    Call this function to enable hooks of Signal
    """
    hyperspy.signal.BaseSignal = HookedSignal


def dehook_signal():
    """
    Call this function to remove hooks from Signal
    """
    hyperspy.signal.BaseSignal = orig_signal

_plotting_cbs = hyperspy.events.Event(
    """Event that triggers right before BaseSignal.plot()
    """,
    ('signal',))
_plotted_cbs = hyperspy.events.Event(
    """Event that triggers right after BaseSignal.plot()
    """,
    ('signal',))


def _on_plotting(signal):
    _plotting_cbs.trigger(signal=signal)


def _on_plotted(signal):
    _plotted_cbs.trigger(signal=signal)


def connect_plotting(callback):
    """
    Call to subscribe to Signal plot events. 'callback' is called on the event.
    The Signal being plotted is passed as the first argument, then follows the
    arguments passed to plot(). Plotting event are called just before the plot
    call is executed.
    """
    _plotting_cbs.connect(callback)


def disconnect_plotting(callback):
    """
    Disconnect callback from subscription.
    """
    _plotting_cbs.disconnect(callback)


def connect_plotted(callback, userdata=None):
    """
    Call to subscribe to Signal plot events. 'callback' is called on the event.
    The Signal being plotted is passed as the first argument, then follows the
    arguments passed to plot(). If userdata is not None is is passed as a
    keyword argument, otherwise, it is left out. Plotted event are called just
    after the plot call is executed.
    """
    _plotted_cbs.connect(callback)


def disconnect_plotted(callback):
    """
    Disconnect callback from subscription.
    """
    _plotted_cbs.pop(callback)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Dec 07 10:30:08 2014

@author: Vidar Tonaas Fauske
"""

import os
import numpy as np
from matplotlib.widgets import SpanSelector

from hyperspy.components1d import Gaussian
from hyperspy.components1d import GaussianHF

from .figuretool import FigureTool
from hyperspyui.util import crosshair_cursor

GaussTypes = (Gaussian, GaussianHF)


class GaussianTool(FigureTool):

    def __init__(self, windows=None):
        super(GaussianTool, self).__init__(windows)
        self.dragging = False
        self.drag_data = None
        self.span = None
        self._old_plot_comp = {}

    def get_name(self):
        return "Gaussian tool"

    def get_category(self):
        return 'Spectrum'

    def get_icon(self):
        return os.path.dirname(__file__) + '/../images/gaussian.svg'

    def is_selectable(self):
        return True

    def make_cursor(self):
        return crosshair_cursor()

    def _wire_wrapper(self, wrapper):
        if wrapper is None:
            return
        m = wrapper.model
        if m._plot.is_active:
            self._old_plot_comp[wrapper] = m._plot_components
            m.enable_plot_components()
            for c in m:
                if isinstance(c, GaussTypes):
                    c._model_plot_line.line.set_picker(True)

    def _unwire_wrapper(self, wrapper):
        if wrapper is None:
            return
        m = wrapper.model
        if wrapper in self._old_plot_comp:
            for c in m:
                if isinstance(c, GaussTypes):
                    c._model_plot_line.line.set_picker(False)
            if not self._old_plot_comp[wrapper]:
                m.disable_plot_components()

    def on_pick(self, event):
        if event.mouseevent.inaxes is None:
            return
        ax = event.mouseevent.inaxes
        if event.mouseevent.button == 3:
            window = ax.figure.canvas.parent()
            mw = window.property('hyperspyUI.ModelWrapper')
            if mw is not None:
                for c in mw.model:
                    line = c._model_plot_line.line
                    if event.artist == line and \
                            isinstance(c, GaussTypes):
                        mw.remove_component(c)

    def connect_windows(self, windows):
        super(GaussianTool, self).connect_windows(windows)
        windows = self._iter_windows(windows)
        for w in windows:
            mw = w.property('hyperspyUI.ModelWrapper')
            self._wire_wrapper(mw)

    def disconnect_windows(self, windows):
        super(GaussianTool, self).disconnect_windows(windows)
        windows = self._iter_windows(windows)
        for w in windows:
            mw = w.property('hyperspyUI.ModelWrapper')
            self._unwire_wrapper(mw)

    def on_mousedown(self, event):
        if event.inaxes is None:
            return
        ax = event.inaxes
        f = ax.figure
        x, y = event.xdata, event.ydata

        if event.button == 1 and event.dblclick:
            # Add Gaussian here!
            window = f.canvas.parent()
            mw = window.property('hyperspyUI.ModelWrapper')
            if mw is None:
                sw = window.property('hyperspyUI.SignalWrapper')
                mw = sw.make_model()
                self._wire_wrapper(mw)
            m = mw.model
            i = m.axis.value2index(x)
            h = m.signal()[i] - m()[i]
            if m.signal.metadata.Signal.binned:
                h /= m.axis.scale
            g = GaussianHF(height=h * np.sqrt(2 * np.pi), centre=x)
            g.height.free = False
            g.centre.free = False
            mw.add_component(g)
            g._model_plot_line.line.set_picker(True)
            m.fit_component(g, signal_range=None, estimate_parameters=False)
            g.height.free = True
            g.centre.free = True
        else:
            self.dragging = True
            self.drag_data = [x, y, event.button, ax, event]

    def on_spanselect(self, x0, x1):
        self.span.disconnect_events()
        self.span = None
        ax = self.drag_data[3]
        window = ax.figure.canvas.parent()
        mw = window.property('hyperspyUI.ModelWrapper')
        if mw is None:
            sw = window.property('hyperspyUI.SignalWrapper')
            mw = sw.make_model()
            self._wire_wrapper(mw)
        m = mw.model

        g = GaussianHF()
        mw.add_component(g)
        g._model_plot_line.line.set_picker(True)
        m.fit_component(g, signal_range=(x0, x1))
        i = m.axis.value2index(g.centre.value)
        g.active = False
        h = m.signal()[i] - m(onlyactive=True)[i]
        g.active = True
        if m.signal.metadata.Signal.binned:
            h /= m.axis.scale
        g.height.value = h
        g.height.free = False
        g.centre.free = False
        m.fit_component(g, signal_range=(x0, x1), estimate_parameters=False)
        g.height.free = True
        g.centre.free = True

        if self.drag_data is None:
            return
        self.drag_data = None

    def on_mousemove(self, event):
        if not self.dragging or self.drag_data is None:
            return

        ax = self.drag_data[3]
        if self.span is None:
            self.dragging = False
            invtrans = ax.transData.inverted()
            minspan = 3 * np.abs(invtrans.transform((1, 0)) -
                                 invtrans.transform((0, 0)))[0]
            self.span = SpanSelector(ax, self.on_spanselect, 'horizontal',
                                     minspan=minspan)
            event2 = self.drag_data[4]
            self.span.press(event2)
            self.span.onmove(event)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.

from functools import partial
from collections import OrderedDict

from qtpy import QtCore, QtWidgets
from qtpy.QtWidgets import QDialogButtonBox
from qtpy.QtCore import Qt, QModelIndex

from hyperspyui.plugins.plugin import Plugin
from hyperspyui.widgets.extendedqwidgets import ExToolWindow
from hyperspy.misc.utils import DictionaryTreeBrowser


def tr(text):
    return QtCore.QCoreApplication.translate("MetadataEditor", text)


class MetadataEditor(Plugin):
    name = "Metadata Editor"

    def __init__(self, *args, **kwargs):
        super(MetadataEditor, self).__init__(*args, **kwargs)
        self.editors = {}

    def create_actions(self):
        self.add_action(self.name + '.view_metadata',
                        tr("View metadata"),
                        self.edit_metadata,
                        tip=tr("View/edit signal meatadata"))

    def create_menu(self):
        self.add_menuitem(
            'Signal', self.ui.actions[self.name + '.view_metadata'])

    def _on_close(self, signal):
        self.editors.pop(signal)

    def edit_metadata(self, signal=None):
        if signal is None:
            signal = self.ui.get_selected_signal()
        if signal in self.editors:
            self.editors[signal].show()
        else:
            diag = ExToolWindow(parent=self.ui)
            sw = self.ui.lut_signalwrapper[signal]
            diag.setWindowTitle(tr("Metadata of %s") % sw.name)
            vbox = QtWidgets.QVBoxLayout()
            tw = MetadataTreeWidget(signal)
            vbox.addWidget(tw)
            btns = QDialogButtonBox(QDialogButtonBox.Ok)
            btns.accepted.connect(diag.accept)
            vbox.addWidget(btns)
            diag.setLayout(vbox)
            diag.show()
            diag.finished.connect(partial(self._on_close, signal))
            self.editors[signal] = diag


class MetadataTreeWidget(QtWidgets.QTreeView):

    def __init__(self, signal, parent=None):
        super(MetadataTreeWidget, self).__init__(parent)
        self.signal = signal
        self._model = MetadataModel(signal)
        self.setModel(self._model)
        self.setAlternatingRowColors(True)
        self.setColumnWidth(0, 300)
        self.setColumnWidth(0, 300)
#        self.setEditTriggers(
#            QAbstractItemView.DoubleClicked |
#            QAbstractItemView.SelectedClicked |
#            QAbstractItemView.EditKeyPressed)
        # Expand 'metadata' by default
        self.expand(self._model.index(0, 0))

    def sizeHint(self):
        hint = super(MetadataTreeWidget, self).sizeHint()
        hint.setWidth(600)
        return hint


# Parts of the following code has the following licence:
# Copyright 2015 Hardcoded Software (http://www.hardcoded.net)
#
# This software is licensed under the "GPLv3" License as described in the
# "LICENSE" file, which should be included with this package. The terms are
# also available at http://www.gnu.org/licenses/gpl-3.0.html

class MetadataNode:
    def __init__(self, model, parent, name, ref, row):
        self._subnodes = None
        self.ref = ref
        self.model = model
        self.parent = parent
        self.row = row
        self.name = name

    def _getChildren(self):
        if isinstance(self.ref, DictionaryTreeBrowser):
            return OrderedDict(self.ref)
        else:
            return {}

    def _createNode(self, name, ref, row):
        return MetadataNode(self.model, self, name, ref, row)

    @property
    def subnodes(self):
        if self._subnodes is None:
            children = self._getChildren()
            self._subnodes = []
            for index, (name, child) in enumerate(children.items()):
                node = self._createNode(name, child, index)
                self._subnodes.append(node)
        return self._subnodes

    @property
    def index(self):
        return self.model.createIndex(self.row, 0, self)

    def edit_value(self, value):
        if self.parent and isinstance(self.ref, (int, str, bool, float)):
            # Immutable type, make sure we change value in parent!
            self.parent.ref[self.name] = value
            self.ref = value
            return True
        return False


class MetadataModel(QtCore.QAbstractItemModel):

    def __init__(self, signal, parent=None):
        super(MetadataModel, self).__init__(parent=parent)
        self.nodes = [
            MetadataNode(self, None, 'metadata', signal.metadata, 0),
            MetadataNode(self, None, 'original_metadata',
                         signal.original_metadata, 1)]

    def index(self, row, column, parent=None):
        if parent and parent.isValid():
            node = parent.internalPointer()
            ref = node.subnodes[row]
        else:
            ref = self.nodes[row]
        return self.createIndex(row, column, ref)

    def parent(self, index):
        if index is None or not index.isValid():
            return QModelIndex()
        node = index.internalPointer()
        if node.parent is None:
            return QModelIndex()
        else:
            return self.createIndex(node.parent.row, 0, node.parent)

    def rowCount(self, parent=None):
        if parent is None or not parent.isValid():
            return len(self.nodes)
        else:
            return len(parent.internalPointer().subnodes)

    def columnCount(self, parent=None):
        return 2

    def flags(self, index):
        flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled
        if index and index.isValid() and index.column() == 1:
            node = index.internalPointer()
            if not isinstance(node.ref, DictionaryTreeBrowser):
                if isinstance(node.ref, bool):
                    flags |= Qt.ItemIsUserCheckable
                elif isinstance(node.ref, (str, int, float)):
                    flags |= Qt.ItemIsEditable
        return flags

    def data(self, index, role):
        if index is None or not index.isValid():
            return None
        node = index.internalPointer()
        if role in (Qt.DisplayRole, Qt.EditRole):
            if index.column() == 0:
                return node.name
            if (index.column() == 1 and
                    not isinstance(node.ref, DictionaryTreeBrowser)):
                return str(node.ref)
        elif role == Qt.CheckStateRole:
            if index.column() == 1 and isinstance(node.ref, bool):
                return Qt.Checked if node.ref else Qt.Unchecked
        elif role == Qt.TextAlignmentRole:
            return Qt.AlignLeft | Qt.AlignTop
        return None

    def setData(self, index, value, role=Qt.EditRole):
        if index is None or not index.isValid() or index.column() != 1:
            return False
        node = index.internalPointer()
        value_changed = False
        input_ok = False
        if role == Qt.CheckStateRole:
            if isinstance(node.ref, bool):
                value = value == Qt.Checked
                input_ok = True
                if node.ref != value:
                    node.edit_value(value)
                    value_changed = True
        elif role == Qt.EditRole:
            if isinstance(node.ref, (str, int, float)):
                try:
                    value = type(node.ref)(value)
                except ValueError:
                    pass
                else:
                    input_ok = True
                    if node.ref != value:
                        node.edit_value(value)
                        value_changed = True
        else:
            return super(MetadataModel, self).setData(index, value, role)
        if value_changed:
            self.dataChanged.emit(index, index)
        return input_ok

    def headerData(self, column, orientation, role):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if column == 0:
                return 'Key'
            elif column == 1:
                return 'Value'
        return None

    def reset(self):
        super().beginResetModel()
        self.invalidate()
        self._ref2node = {}
        self._dummyNodes = set()
        super().endResetModel()



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Based off of:

ZetCode PyQt4 tutorial

This program creates a skeleton of
a classic GUI application with a menubar,
toolbar, statusbar and a central widget.

author: Jan Bodnar
website: zetcode.com
last edited: September 2011
"""

import sys
from qtpy import QtGui, QtWidgets

import tempfile
import glob
import os


class Example(QtWidgets.QMainWindow):

    def __init__(self):
        super(Example, self).__init__()

        self.initUI()

    def replaceSvgColor(self, filename, oldColor, newColor):
        '''
        Generate and return a temp svg filename with oldColor replaced
        by newColor (color names are in HTML format '#XXXXXX')
        '''
        svg_file = open(filename, 'r')
        svg_cnt = svg_file.read()
        temp = tempfile.NamedTemporaryFile(delete=False)
        temp.write(svg_cnt.replace(oldColor, newColor).encode('utf-8'))
        temp.close()

        return temp.name

    def initUI(self):
        stack = QtWidgets.QVBoxLayout()
        widget = QtWidgets.QWidget()

        color_list = []
        color_list.append(('QtGui.QPalette.Window',
                           self.palette().color(QtGui.QPalette.Window)))
        color_list.append(('QtGui.QPalette.Background',
                           self.palette().color(QtGui.QPalette.Background)))
        color_list.append(('QtGui.QPalette.WindowText',
                           self.palette().color(QtGui.QPalette.WindowText)))
        color_list.append(('QtGui.QPalette.Foreground',
                           self.palette().color(QtGui.QPalette.Foreground)))
        color_list.append(('QtGui.QPalette.Base',
                           self.palette().color(QtGui.QPalette.Base)))
        color_list.append(('QtGui.QPalette.AlternateBase',
                           self.palette().color(QtGui.QPalette.AlternateBase)))
        color_list.append(('QtGui.QPalette.ToolTipBase',
                           self.palette().color(QtGui.QPalette.ToolTipBase)))
        color_list.append(('QtGui.QPalette.ToolTipText',
                           self.palette().color(QtGui.QPalette.ToolTipText)))
        color_list.append(('QtGui.QPalette.Text',
                           self.palette().color(QtGui.QPalette.Text)))
        color_list.append(('QtGui.QPalette.Button',
                           self.palette().color(QtGui.QPalette.Button)))
        color_list.append(('QtGui.QPalette.ButtonText',
                           self.palette().color(QtGui.QPalette.ButtonText)))
        color_list.append(('QtGui.QPalette.BrightText',
                           self.palette().color(QtGui.QPalette.BrightText)))

        textEdit = QtWidgets.QTextEdit()
        for n, c in color_list:
            textEdit.append(n + ':\t' + str(c.name()))
        stack.addWidget(textEdit)

        # Change this to wherever your icons are stored
        icon_files = glob.glob("*.svg")

        widget.setLayout(stack)
        self.setCentralWidget(widget)

        toolbar1 = self.addToolBar('Bar1')
        toolbar2 = self.addToolBar('Bar2')
        toolbar3 = self.addToolBar('Bar3')

        ico_list = [''] * len(icon_files)
        action_list = [''] * len(icon_files)
        for i, f in enumerate(icon_files):
            ico_list[i] = QtGui.QIcon(self.replaceSvgColor(f,
                                                           '#000000',
                                                           'red'))
            action_list[i] = QtWidgets.QAction(ico_list[i], 'Exit', self)
            action_list[i].setStatusTip(os.path.basename(f))

            # add to toolbars
            if i/float(len(icon_files)) <= 0.33:
                toolbar1.addAction(action_list[i])
            elif i/float(len(icon_files)) <= 0.66:
                toolbar2.addAction(action_list[i])
            else:
                toolbar3.addAction(action_list[i])

        # shortcut to close ('X' icon in position 8)
        action_list[8].setShortcut('Ctrl+Q')
        action_list[8].triggered.connect(self.close)

        self.insertToolBarBreak(toolbar2)
        self.insertToolBarBreak(toolbar3)

        self.statusBar()

        menubar = self.menuBar()
        fileMenu = menubar.addMenu('&File')
        fileMenu.addAction(action_list[8])

        self.setGeometry(50, 30, 650, 350)
        self.setWindowTitle('Icon color testing')
        self.show()

        self.statusBar().showMessage("Ready")


def main():
    app = QtWidgets.QApplication(sys.argv)
    _ = Example()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri Oct 24 16:46:35 2014

@author: Vidar Tonaas Fauske
"""

from collections import OrderedDict
from functools import partial
import argparse
import os
import sys
import json
import webbrowser

import numpy as np

# Hyperspy uses traitsui, set proper backend
from traits.etsconfig.api import ETSConfig
try:
    ETSConfig.toolkit = 'qt4'
except ValueError:
    if 'sphinx' not in sys.modules:
        raise

from qtpy import QtGui, QtCore, QtWidgets
from qtpy.QtCore import Qt

from hyperspyui.mainwindowhyperspy import MainWindowHyperspy, tr
from hyperspyui.util import create_add_component_actions, win2sig, dict_rlu
from hyperspyui.widgets.contrastwidget import ContrastWidget
from hyperspyui.widgets.editorwidget import EditorWidget
from hyperspyui.widgets.pluginmanagerwidget import PluginManagerWidget
from hyperspyui.widgets.pickxsignals import PickXSignalsWidget
from hyperspyui.log import logger
import hyperspyui.tools


class MainWindow(MainWindowHyperspy):

    """
    Main window of the application. Top layer in application stack. Is
    responsible for adding default actions, and filling the menus and toolbars.
    Also creates the default widgets. Any button-actions should also be
    accessible as a slot, such that other things can connect into it, and so
    that it is accessible from the console's 'ui' variable.
    """

    load_complete = QtCore.Signal()

    def __init__(self, splash, parent=None, argv=None):
        self.splash = splash

        # State variables
        self.signal_type_ag = None
        self.signal_datatype_ag = None
        self._plugin_manager_widget = None

        self._load_signal_types()

        super(MainWindow, self).__init__(parent)

        # Set window icon
        self.setWindowIcon(QtGui.QIcon(os.path.join(os.path.dirname(__file__),
                                                    'images', 'hyperspy.svg')))

        # Parse any command line options
        self.parse_args(argv)

        # All good!
        self.set_status("Ready")

        # Redirect streams (wait until the end to not affect during load)
        self.settings.set_default('output_to_console', False)
        if self.settings['output_to_console', bool]:
            self._old_stdout = sys.stdout
            self._old_stderr = sys.stdout
            sys.stdout = self.console.kernel.stdout
            sys.stderr = self.console.kernel.stderr
        else:
            self._old_stdout = self._old_stderr = None
        logger.info("Main window loaded!")

        # Set the UI in front of other applications
        self.show()
        self.raise_()
        # Workaround to bring the floating console to the front with pyqt5
        if self._console_dock.isFloating():
            self._console_dock.setFloating(True)

    def set_splash(self, message):
        """Set splash message"""
        if self.splash is None:
            return
        if message:
            logger.debug(message)
        self.splash.show()
        self.splash.showMessage(message, Qt.AlignBottom | Qt.AlignCenter |
                                Qt.AlignAbsolute, QtGui.QColor(Qt.white))
        QtWidgets.QApplication.processEvents()

    def _load_signal_types(self):
        self.set_splash('Loading HyperSpy signals...')
        import hyperspy.signals
        self.signal_types = OrderedDict(
            [('Signal', hyperspy.signal.BaseSignal),
             ('1D Signal', hyperspy.signals.Signal1D),
             ('2D Signal', hyperspy.signals.Signal2D),
             ('EELS', hyperspy.signals.EELSSpectrum),
             ('EDS SEM', hyperspy.signals.EDSSEMSpectrum),
             ('EDS TEM', hyperspy.signals.EDSTEMSpectrum),
             ('Complex Signal 1D', hyperspy.signals.ComplexSignal1D),
             ('Complex Signal 2D', hyperspy.signals.ComplexSignal2D),
             ('Dielectric Function', hyperspy.signals.DielectricFunction),
             ])

    def handleSecondInstance(self, argv):
        """
        A second instance was launched and suppressed. Process the arguments
        that were passed to the new instance.
        """
        super(MainWindow, self).handleSecondInstance(argv)
        argv = json.loads(argv)
        self.parse_args(argv)

    def parse_args(self, argv=None):
        """
        Parse command line arguments, either from sys.argv, or from parameter
        'argv'.
        """
        parser = argparse.ArgumentParser(
            description=QtCore.QCoreApplication.applicationName() +
            " " + QtCore.QCoreApplication.applicationVersion())
        parser.add_argument('files', metavar='file', type=str, nargs='*',
                            help='data file to open.')
        if argv is None:
            args = parser.parse_args()
        else:
            args = parser.parse_args(argv)
        files = args.files

        if len(files) > 0:
            self.load(files)

    def create_default_actions(self):
        super(MainWindow, self).create_default_actions()

        # Files:
        self.add_action('open', "&Open", self.load,
                        shortcut=QtGui.QKeySequence.Open,
                        icon='open.svg',
                        tip="Open existing file(s)")
        self.add_action('open_stack', "Open S&tack", self.load_stack,
                        tip="Open files and combine into one signal (stacked)")
        self.add_action('close', "&Close", self.close_signal,
                        shortcut=QtGui.QKeySequence.Close,
                        icon='close_window.svg',
                        selection_callback=self.select_signal,
                        tip="Close the selected signal(s)")
        self.add_action('new_editor', "&New editor", self.new_editor,
                        shortcut=QtGui.QKeySequence.New,
                        tip="Opens a new code editor")

        close_all_key = QtGui.QKeySequence(Qt.CTRL + Qt.ALT + Qt.Key_F4,
                                           Qt.CTRL + Qt.ALT + Qt.Key_W)
        self.add_action('close_all', "&Close All", self.close_all_signals,
                        shortcut=close_all_key,
                        icon='close_windows.svg',
                        tip="Close all signals")
        self.add_action('exit', "E&xit", self.close,
                        shortcut=QtGui.QKeySequence.Quit,
                        tip="Exits the application")

        # I/O:
        self.add_action('save', "&Save", self.save,
                        shortcut=QtGui.QKeySequence.Save,
                        icon='save.svg',
                        selection_callback=self.select_signal,
                        tip="Save the selected signal(s)")
        self.add_action('save_fig', "Save &figure", self.save_figure,
                        tip="Save the active figure")

        self.add_action('add_model', "Create Model", self.add_model,
                        selection_callback=self.select_signal,
                        tip="Create a model for the selected signal")

        # Settings:
        self.add_action('plugin_manager', "Plugin manager",
                        self.show_plugin_manager,
                        tip="Show the plugin manager")
        self.add_action('hspy_settings', "HyperSpy settings",
                        self.edit_hspy_settings,
                        tip="Edit the HyperSpy package settings")
        self.add_action('edit_settings', "Edit settings", self.edit_settings,
                        tip="Edit the application and plugins settings")

        # Help:
        self.add_action('documentation', "Documentation",
                        self.open_documentation,
                        tip="Open the HyperSpyUI documentation in a browser.")

        # --- Add signal type selection actions ---
        signal_type_ag = QtWidgets.QActionGroup(self)
        signal_type_ag.setExclusive(True)
        for st in self.signal_types.keys():
            f = partial(self.set_signal_type, st)
            st_ac = self.add_action('signal_type_' + st, st, f)
            st_ac.setCheckable(True)
            signal_type_ag.addAction(st_ac)
        self.signal_type_ag = signal_type_ag

        # --- Add signal data type selection actions ---
        signal_datatype_ag = QtWidgets.QActionGroup(self)
        signal_datatype_ag.setExclusive(True)
        import numpy as np
        for t in [np.bool, np.bool8, np.byte, np.complex, np.complex64,
                  np.complex128, np.float, np.float16, np.float32, np.float64,
                  np.int, np.int8, np.int16, np.int32, np.int64, np.long,
                  np.uint, np.uint8, np.uint16, np.uint32, np.uint64, 'Custom'
                  ]:
            f = partial(self.set_signal_dtype, t)
            if isinstance(t, str):
                st = t.lower()
            else:
                st = t.__name__

            sdt_ac = self.add_action('signal_data_type_' + st, st, f)
            sdt_ac.setCheckable(True)
            signal_datatype_ag.addAction(sdt_ac)
        self.signal_datatype_ag = signal_datatype_ag

        # Start disabled until a valid figure is selected
        self.signal_type_ag.setEnabled(False)
        self.signal_datatype_ag.setEnabled(False)

        # --- Add "add component" actions ---
        comp_actions = create_add_component_actions(self, self.make_component)
        self.comp_actions = []
        for ac_name, ac in comp_actions.items():
            self.actions[ac_name] = ac
            self.comp_actions.append(ac_name)
            self._action_selection_cbs[ac_name] = self._check_add_component_ok
            ac.setEnabled(False)

    def create_menu(self):
        mb = self.menuBar()

        # File menu (I/O)
        self.menus['File'] = mb.addMenu(tr("&File"))
        self.add_menuitem('File', self.actions['open'])
        self.add_menuitem('File', self.actions['open_stack'])
        self.add_menuitem('File', self.actions['close'])
        self.add_menuitem('File', self.actions['save'])
        self.add_menuitem('File', self.actions['save_fig'])
        self.add_menuitem('File', self.actions['new_editor'])

        # Signal menu
        self.menus['Signal'] = mb.addMenu(tr("&Signal"))
        stm = self.menus['Signal'].addMenu(tr("Signal type"))
        for ac in self.signal_type_ag.actions():
            stm.addAction(ac)
        stm = self.menus['Signal'].addMenu(tr("Signal data type"))
        for ac in self.signal_datatype_ag.actions():
            stm.addAction(ac)

        # Model menu
        self.menus['Model'] = mb.addMenu(tr("&Model"))
        self.add_menuitem('Model', self.actions['add_model'])
        self.modelmenu_sep1 = self.menus['Model'].addSeparator()

        componentmenu = self.menus['Model'].addMenu(tr("&Add Component"))
        for acname in self.comp_actions:
            componentmenu.addAction(self.actions[acname])

        # Create Windows menu
        super(MainWindow, self).create_menu()

        self.menus['File'].addSeparator()
        self.add_menuitem('File', self.actions['close_all'])
        self.add_menuitem('File', self.actions['exit'])

        # Ensure settings menu next to last
        if 'Settings' in self.menus:
            m = self.menus['Settings']
            self.menuBar().removeAction(m.menuAction())
            self.menuBar().insertMenu(self.windowmenu.menuAction(), m)

        self.add_menuitem('Settings', self.actions['plugin_manager'])
        self.add_menuitem('Settings', self.actions['reset_layout'])
        self.add_menuitem('Settings', self.actions['hspy_settings'])
        self.add_menuitem('Settings', self.actions['edit_settings'])

        # Create Help menu, so it is searchable on Mac
        self.menus['Help'] = mb.addMenu(tr("&Help"))
        self.add_menuitem('Help', self.actions['documentation'])

    def create_tools(self):
        super(MainWindow, self).create_tools()
        for tool_type in hyperspyui.tools.default_tools:
            self.add_tool(tool_type)

    def create_toolbars(self):
        self.add_toolbar_button("Files", self.actions['open'])
        self.add_toolbar_button("Files", self.actions['close'])
        self.add_toolbar_button("Files", self.actions['save'])

        super(MainWindow, self).create_toolbars()

    def create_widgetbar(self):
        super(MainWindow, self).create_widgetbar()

        self.add_widget(ContrastWidget(self, self))

    # ---------------------------------------
    # Events
    # ---------------------------------------

    def _check_add_component_ok(self, win, action):
        s = win2sig(win, self.signals, self._plotting_signal)
        if s is None:
            action.setEnabled(False)
        else:
            action.setEnabled(True)

    def on_subwin_activated(self, mdi_figure):
        super(MainWindow, self).on_subwin_activated(mdi_figure)
        s = win2sig(mdi_figure, self.signals, self._plotting_signal)
        if s is None:
            for ac in self.signal_type_ag.actions():
                ac.setChecked(False)
            for ac in self.signal_datatype_ag.actions():
                ac.setChecked(False)
            self.signal_type_ag.setEnabled(False)
            self.signal_datatype_ag.setEnabled(False)
        else:
            t = type(s.signal)
            key = 'signal_type_' + dict_rlu(self.signal_types, t)
            self.actions[key].setChecked(True)
            key2 = 'signal_data_type_' + s.signal.data.dtype.type.__name__
            if key2 in self.actions:
                self.actions[key2].setChecked(True)
            else:
                self.actions['signal_data_type_custom'].setChecked(True)
            self.signal_type_ag.setEnabled(True)
            self.signal_datatype_ag.setEnabled(True)

    def on_settings_changed(self):
        # Redirect streams (wait until the end to not affect during load)
        super(MainWindow, self).on_settings_changed()
        if self.settings['output_to_console', bool]:
            if self._old_stdout is None:
                self._old_stdout = sys.stdout
                self._old_stderr = sys.stderr
                sys.stdout = self.console.kernel.stdout
                sys.stderr = self.console.kernel.stderr
        else:
            if self._old_stdout is not None:
                sys.stdout = self._old_stdout
                sys.stderr = self._old_stderr

    # ---------------------------------------
    # Slots
    # ---------------------------------------

    def select_x_signals(self, x, titles=None, wrap_col=4):
        """
        Displays a blocking dialog prompting the user to select 'x' signals.
        Over each selection box, the title as defined by 'titles' is displayed.
        """
        w = PickXSignalsWidget(self.signals, x, parent=self,
                               titles=titles, wrap_col=wrap_col)
        diag = self.show_okcancel_dialog("Select signals", w, True)
        signals = None
        if diag.result() == QtWidgets.QDialog.Accepted:
            signals = w.get_selected()
        w.unbind()
        return signals

    def show_plugin_manager(self):
        if self._plugin_manager_widget is None:
            self._plugin_manager_widget = PluginManagerWidget(
                self.plugin_manager, self)
        self._plugin_manager_widget.show()

    def close_signal(self, uisignals=None):
        uisignals = self.get_selected_wrappers()
        for s in uisignals:
            s.close()

    def close_all_signals(self):
        while len(self.signals) > 0:
            s = self.signals.pop()
            s.close()

    def new_editor(self):
        e = EditorWidget(self, self)
        self.editors.append(e)
        e.show()

    def set_signal_type(self, signal_type, signal=None):
        """
        Changes the signal type using a combination of hyperspy.Signal.:
         * set_signal_type()
         * set_signal_origin()
         * and by converting with as_signal1D() and as_signal2D()
        """
        # The list of signal types is maintained as a list here and in
        # self.signal_types, as the names can be adapted, and since they need
        # to be diferentiated based on behavior either way.
        if signal is None:
            signal = self.get_selected_wrapper()
        self.record_code("signal = ui.get_selected_signal()")

        # Sanity check
        if signal_type not in list(self.signal_types.keys()):
            raise ValueError()

        self.setUpdatesEnabled(False)
        try:
            import hyperspy.signals
            if signal_type in ['2D Signal', 'Complex Signal 2D']:
                if not isinstance(signal.signal,
                                  (hyperspy.signals.Signal2D,
                                   hyperspy.signals.ComplexSignal2D)):
                    signal.as_signal2D((0, 1))
                    self.record_code("signal = signal.as_signal2D((0, 1))")
            else:
                if isinstance(signal.signal,
                              (hyperspy.signals.Signal2D,
                               hyperspy.signals.ComplexSignal2D)):
                    signal.as_signal1D(0)
                    self.record_code("signal = signal.as_signal1D(0)")

            if signal_type in ['EELS', 'EDS SEM', 'EDS TEM']:
                underscored = signal_type.replace(" ", "_")
                signal.signal.set_signal_type(underscored)
                self.record_code("signal.set_signal_type('%s')" % underscored)

            signal.plot()
        finally:
            self.setUpdatesEnabled(True)

    def set_signal_dtype(self, data_type, signal=None, clip=False):
        if signal is None:
            signal = self.get_selected_signal()
        self.record_code("signal = ui.get_selected_signal()")
        if isinstance(data_type, str) and data_type.lower() == 'custom':
            return    # TODO: Show dialog and prompt
        if not clip:
            old_type = signal.data.dtype
            if np.issubdtype(data_type, np.integer):
                info = np.iinfo(data_type)
            elif np.issubdtype(data_type, np.float):
                info = np.finfo(data_type)
            if np.issubdtype(old_type, np.integer):
                old_info = np.iinfo(old_type)
            elif np.issubdtype(old_type, np.float):
                old_info = np.finfo(old_type)
            if old_info.max > info.max:
                signal.data *= float(info.max) / np.nanmax(signal.data)
                self.record_code("signal.data *= %f / np.nanmax(signal.data)" %
                                 float(info.max))
        signal.change_dtype(data_type)
        dts = data_type.__name__
        if data_type.__module__ == 'numpy':
            dts = 'np.' + dts
        self.record_code("signal.change_dtype(%s)" % dts)

    def open_documentation(self):
        webbrowser.open('http://hyperspy.org/hyperspyUI/')



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri Nov 21 22:22:33 2014

@author: Vidar Tonaas Fauske
"""

from functools import partial

from qtpy import QtCore, QtWidgets

import hyperspy.signals
from hyperspy.misc.elements import elements as elements_db

from hyperspyui.widgets.extendedqwidgets import FigureWidget, ExClickLabel
from hyperspyui.widgets.periodictable import PeriodicTableWidget
from hyperspyui.util import win2sig, block_signals


def tr(text):
    return QtCore.QCoreApplication.translate("ElementPickerWidget", text)

edstypes = (hyperspy.signals.EDSSEMSpectrum, hyperspy.signals.EDSTEMSpectrum)


class ElementPickerWidget(FigureWidget):

    """
    Tool window for picking elements of an interactive periodic table.
    Takes a signal in the constructor, and a parent control.
    """
    element_toggled = QtCore.Signal(str)

    def __init__(self, main_window, parent):
        super(ElementPickerWidget, self).__init__(main_window, parent)
        self.signal = None
        self.create_controls()
        self.table.element_toggled.connect(self._toggle_element)
        self.set_signal(self.signal)
        self._only_lines = ['Ka', 'La', 'Ma',
                            'Kb', 'Lb1', 'Mb']

    def _on_figure_change(self, figure):
        super(ElementPickerWidget, self)._on_figure_change(figure)
        signal = win2sig(figure, plotting_signal=self.ui._plotting_signal)
        self.set_signal(signal)

    def set_signal(self, signal):
        self.signal = signal
        self.setWindowTitle(tr("Element picker"))

        if self.isEDS():
            self.map_btn.show()
        else:
            self.map_btn.hide()

        if signal is None or signal.signal is None:
            self.set_enabled(False)
            return
        else:
            self.set_enabled(True)

        # Enable markers if plot has any
        with block_signals(self.chk_markers):
            markers = (hasattr(signal.signal, '_xray_markers') and
                       bool(signal.signal._xray_markers))
            self.chk_markers.setChecked(markers)

        # Make sure we have the Sample node, and Sample.elements
        if not hasattr(signal.signal.metadata, 'Sample'):
            signal.signal.metadata.add_node('Sample')
        if not hasattr(signal.signal.metadata.Sample, 'elements'):
            signal.signal.metadata.Sample.elements = []

        self._set_elements(signal.signal.metadata.Sample.elements)

        # Disable elements which hyperspy does not accept
        hsyp_elem = list(elements_db.keys())
        for w in self.table.children():
            if isinstance(w, ExClickLabel):
                elem = w.text()
                if elem not in hsyp_elem:
                    self.table.disable_element(elem)
                else:
                    self.table.enable_element(elem)

    def set_enabled(self, value):
        self.setEnabled(value)

    @property
    def markers(self):
        return self.chk_markers.isChecked()

    def isEDS(self):
        if self.signal is None or self.signal.signal is None:
            return False
        return isinstance(self.signal.signal, edstypes)

    def isEELS(self):
        if self.signal is None or self.signal.signal is None:
            return False
        return isinstance(self.signal.signal, hyperspy.signals.EELSSpectrum)

    def _toggle_element(self, element):
        """
        Makes sure the element is toggled correctly for both EDS and EELS.
        Dependent on hyperspy implementation, as there are currently no
        remove_element functions.
        """
        if self.isEELS():
            self._toggle_element_eels(element)
        elif self.isEDS():
            self._toggle_element_eds(element)

    def _toggle_element_eds(self, element):
        s = self.signal.signal
        lines_added = []
        lines_removed = []
        self.ui.record_code("signal = ui.get_selected_signal()")
        if element in s.metadata.Sample.elements:
            # Element present, we're removing it
            s.metadata.Sample.elements.remove(element)
            self.ui.record_code(
                "signal.metadata.Sample.elements.remove('%s')" % element)
            if 'Sample.xray_lines' in s.metadata:
                for line in reversed(s.metadata.Sample.xray_lines):
                    if line.startswith(element):
                        s.metadata.Sample.xray_lines.remove(line)
                        lines_removed.append(line)
                if len(s.metadata.Sample.xray_lines) < 1:
                    del s.metadata.Sample.xray_lines
            else:
                lines_removed.extend(s._get_lines_from_elements(
                                     [element], only_one=False,
                                     only_lines=self._only_lines))
        else:
            lines_added = s._get_lines_from_elements(
                [element], only_one=False, only_lines=self._only_lines)
            if 'Sample.xray_lines' in s.metadata:
                s.add_lines(lines_added)  # Will also add element
                self.ui.record_code(
                    "signal.add_lines(%s)" % str(lines_added))
            else:
                s.add_elements((element,))
                self.ui.record_code(
                    "signal.add_elements(%s)" % str([element]))
        if self.markers:
            if lines_added:
                s.add_xray_lines_markers(lines_added)
            if lines_removed:
                s.remove_xray_lines_markers(lines_removed)

    def _toggle_element_eels(self, element):
        s = self.signal.signal
        self.ui.record_code("signal = ui.get_selected_signal()")
        if element in s.metadata.Sample.elements:
            s.elements.remove(element)
            s.subshells = set()
            s.add_elements([])  # Will set metadata.Sample.elements
            self.ui.record_code(
                "signal.elements.remove('%s')" % str(element))
            self.ui.record_code(
                "signal.subshells = set()")
            self.ui.record_code(
                "signal.add_elements([])")
        else:
            s.add_elements((element,))
            self.ui.record_code(
                "signal.add_elements(%s)" % str([element]))

    def _toggle_subshell(self, subshell, checked):
        if not self.isEDS():
            return
        s = self.signal.signal
        element, ss = subshell.split('_')

        # Figure out whether element should be toggled
        active, _ = self._get_element_subshells(element)
        if checked:
            any_left = True
        elif ss in active:
            active.remove(ss)
            any_left = len(active) > 0

        # If any(subshells toggled) != element toggled, we should toggle
        # element
        if self.table.toggled[element] != any_left:
            # Update table toggle
            self.table.toggle_element(element)
            # Update signal state
            if not any_left:
                # Remove element
                self._toggle_element(element)

        self.ui.record_code("signal = ui.get_selected_signal()")
        if 'Sample.xray_lines' not in s.metadata and len(active) > 0:
            lines = [element + '_' + a for a in active]
            s.add_lines(lines)
            self.ui.record_code(
                "signal.add_lines(%s)" % str(lines))
            if self.markers:
                if checked:
                    s.add_xray_lines_markers(lines)
                else:
                    s.remove_xray_lines_markers([subshell])
        else:
            if checked:
                s.add_lines([subshell])
                self.ui.record_code(
                    "signal.add_lines(%s)" % str([subshell]))
                if self.markers:
                    s.add_xray_lines_markers([subshell])
            elif 'Sample.xray_lines' in s.metadata:
                if subshell in s.metadata.Sample.xray_lines:
                    s.metadata.Sample.xray_lines.remove(subshell)
                    self.ui.record_code(
                        "signal.metadata.Sample.xray_lines.remove('%s')" %
                        str(subshell))
                    if self.markers:
                        s.remove_xray_lines_markers([subshell])
                # If all lines are disabled, fall back to element defined
                # (Not strictly needed)
                if len(s.metadata.Sample.xray_lines) < 1:
                    del s.metadata.Sample.xray_lines
                    self.ui.record_code(
                        "del signal.metadata.Sample.xray_lines")

    def _set_elements(self, elements):
        """
        Sets the table elements to passed parameter. Does not modify elements
        in signal! That is handled by the _toggle_element* functions
        """
        self.table.set_elements(elements)

    def set_element(self, element, value):
        """Sets the state of element in table and adds/removes in signal if
        necessary
        """
        self.table.set_element(element, value)
        s = self.signal.signal
        if (element in s.metadata.Sample.elements) != value:
            self._toggle_element(element)

    def _on_toggle_markers(self, value):
        """Toggles peak markers on the plot, i.e. adds/removes markers for all
        elements added on signal.
        """
        w = self.signal
        s = self.signal.signal
        if value:
            if self.isEDS():
                w.keep_on_close = True
                s._plot_xray_lines(xray_lines=True)
                w.update_figures()
                w.keep_on_close = False
        else:
            if self.isEDS():
                for m in reversed(s._plot.signal_plot.ax_markers):
                    m.close(render_figure=False)
                if hasattr(s, '_xray_markers'):
                    s._xray_markers.clear()

    def make_map(self):
        """
        Make integrated intensity maps for the defines elements. Currently
        only implemented for EDS signals.
        """
        if self.isEELS():
            pass    # TODO: EELS maps
        elif self.isEDS():
            imgs = self.signal.signal.get_lines_intensity(only_one=False)
            for im in imgs:
                im.plot()

    def _get_element_subshells(self, element, include_pre_edges=False):
        s = self.signal.signal
        subshells = []
        possible_subshells = []
        if self.isEELS():
            subshells[:] = [ss.split('_')[0] for ss in s.subshells]

            Eaxis = s.axes_manager.signal_axes[0].axis
            if not include_pre_edges:
                start_energy = Eaxis[0]
            else:
                start_energy = 0.
            end_energy = Eaxis[-1]
            for shell in elements_db[element][
                    'Atomic_properties']['Binding_energies']:
                if shell[-1] != 'a':
                    if start_energy <= \
                            elements_db[element]['Atomic_properties'][
                            'Binding_energies'][shell]['onset_energy (eV)'] \
                            <= end_energy:
                        possible_subshells.add(shell)
        elif self.isEDS():
            if 'Sample.xray_lines' in s.metadata:
                xray_lines = s.metadata.Sample.xray_lines
                for line in xray_lines:
                    c_element, subshell = line.split("_")
                    if c_element == element:
                        subshells.append(subshell)
            elif ('Sample.elements' in s.metadata and
                  element in s.metadata.Sample.elements):
                xray_lines = s._get_lines_from_elements(
                    [element], only_one=False, only_lines=self._only_lines)
                for line in xray_lines:
                    _, subshell = line.split("_")
                    subshells.append(subshell)
            possible_xray_lines = \
                s._get_lines_from_elements([element], only_one=False,
                                           only_lines=self._only_lines)
            for line in possible_xray_lines:
                _, subshell = line.split("_")
                possible_subshells.append(subshell)

        return (subshells, possible_subshells)

    def element_context(self, widget, point):
        if not self.isEDS():
            return
        cm = QtWidgets.QMenu()
        element = widget.text()
        active, possible = self._get_element_subshells(element)
        for ss in possible:
            key = element + '_' + ss
            ac = cm.addAction(ss)
            ac.setCheckable(True)
            ac.setChecked(ss in active)
            f = partial(self._toggle_subshell, key)
            ac.toggled[bool].connect(f)
        if possible:
            cm.exec_(widget.mapToGlobal(point))

    def create_controls(self):
        """
        Create UI controls.
        """
        self.table = PeriodicTableWidget(self)
        self.table.element_toggled.connect(self.element_toggled)    # Forward
        for w in self.table.children():
            if not isinstance(w, ExClickLabel):
                continue
            w.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            f = partial(self.element_context, w)
            w.customContextMenuRequested[QtCore.QPoint].connect(f)

        self.chk_markers = QtWidgets.QCheckBox(tr("Markers"))
        self.chk_markers.toggled[bool].connect(self._on_toggle_markers)
        self.map_btn = QtWidgets.QPushButton(tr("Map"))
        self.map_btn.clicked.connect(self.make_map)

        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(self.table)

        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(self.chk_markers)
        hbox.addWidget(self.map_btn)
        vbox.addLayout(hbox)

        w = QtWidgets.QWidget()
        w.setLayout(vbox)
        self.setWidget(w)



"""
Copyright (C) 2015 The Qt Company Ltd.
Contact: http://www.qt.io/licensing/

This file is part of the examples of the Qt Toolkit.

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the name of The Qt Company Ltd nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
"""

from qtpy import QtCore, QtWidgets


class FlowLayout(QtWidgets.QLayout):
    def __init__(self, margin=-1, hSpacing=-1, vSpacing=-1, parent=None):
        super(FlowLayout, self).__init__(parent)
        self._hSpace = hSpacing
        self._vSpace = vSpacing

        self._items = []

        self.setContentsMargins(margin, margin, margin, margin)

    def addItem(self, item):
        self._items.append(item)

    def horizontalSpacing(self):
        if self._hSpace >= 0:
            return self._hSpace
        else:
            return self._smartSpacing(
                QtWidgets.QStyle.PM_LayoutHorizontalSpacing)

    def verticalSpacing(self):
        if self._vSpace >= 0:
            return self._vSpace
        else:
            return self._smartSpacing(
                QtWidgets.QStyle.PM_LayoutVerticalSpacing)

    def expandingDirections(self):
        return QtCore.Qt.Orientation(0)

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        return self._doLayout(
            QtCore.QRect(0, 0, width, 0), True)

    def count(self):
        return len(self._items)

    def itemAt(self, index):
        try:
            return self._items[index]
        except IndexError:
            return None

    def minimumSize(self):
        size = QtCore.QSize()
        for item in self._items:
            size = size.expandedTo(item.minimumSize())

        size += QtCore.QSize(2 * self.margin(), 2 * self.margin())
        return size

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self._doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def takeAt(self, index):
        if index >= 0 and index < len(self._items):
            return self._items.pop(index)
        else:
            return 0

    def _doLayout(self, rect, testOnly):

        left, top, right, bottom = self.getContentsMargins()
        effectiveRect = rect.adjusted(+left, +top, -right, -bottom)
        x = effectiveRect.x()
        y = effectiveRect.y()
        lineHeight = 0

        for item in self._items:
            w = item.widget()
            spaceX = self.horizontalSpacing()
            if spaceX == -1:
                spaceX = w.style().layoutSpacing(
                    QtWidgets.QSizePolicy.PushButton,
                    QtWidgets.QSizePolicy.PushButton,
                    QtCore.Qt.Horizontal)
            spaceY = self.verticalSpacing()
            if spaceY == -1:
                spaceY = w.style().layoutSpacing(
                    QtWidgets.QSizePolicy.PushButton,
                    QtWidgets.QSizePolicy.PushButton,
                    QtCore.Qt.Vertical)
            nextX = x + item.sizeHint().width() + spaceX
            if (nextX - spaceX > effectiveRect.right() and
                    lineHeight > 0):
                x = effectiveRect.x()
                y = y + lineHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0

            if not testOnly:
                item.setGeometry(QtCore.QRect(
                    QtCore.QPoint(x, y), item.sizeHint()))

            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())

        return y + lineHeight - rect.y() + bottom

    def _smartSpacing(self, pm):
        parent = self.parent()
        if not parent:
            return -1
        elif parent.isWidgetType():
            return parent.style().pixelMetric(pm, None, parent)
        else:
            return parent.spacing()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Dec 07 03:49:54 2014

@author: Vidar Tonaas Fauske
"""

import os
from matplotlib.pylab import gca

from .figuretool import FigureTool


class HomeTool(FigureTool):

    def __init__(self, windows=None):
        super(HomeTool, self).__init__(windows)

    def get_name(self):
        return "Home tool"

    def get_category(self):
        return 'Plot'

    def get_icon(self):
        return os.path.join(os.path.dirname(__file__), '..', 'images', 'home.svg')

    def single_action(self):
        return self.home

    def home(self, axes=None):
        if axes is None:
            axes = gca()
            if axes is None:
                return
        oldx = axes.get_autoscalex_on()
        oldy = axes.get_autoscaley_on()
        axes.set_xlim(auto=True)
        axes.set_ylim(auto=True)
        axes.autoscale_view(tight=True, scalex=True, scaley=False)
        axes.autoscale_view(tight=None, scalex=False, scaley=True)
        axes.figure.canvas.draw_idle()
        axes.set_xlim(auto=oldx)
        axes.set_ylim(auto=oldy)

    def on_keyup(self, event):
        if event.key == '5':
            self.home()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Mar 01 15:20:55 2015

@author: Vidar Tonaas Fauske
"""

import sys

from hyperspyui.plugins.plugin import Plugin
from hyperspy.utils import stack as stack_signals

from qtpy import QtCore, QtWidgets

from hyperspyui.widgets.axespicker import AxesPickerDialog


def tr(text):
    return QtCore.QCoreApplication.translate("BasicSignalPlugin", text)


class BasicSignalPlugin(Plugin):
    name = "Basic signal tools"

    def __init__(self, *args, **kwargs):
        super(BasicSignalPlugin, self).__init__(*args, **kwargs)
        self.settings.set_default('histogram_bins_method', 'freedman')
        self.settings.set_enum_hint(
            'histogram_bins_method',
            ('knuth', 'scotts', 'freedman', 'blocks', '<integer>'))

    def create_actions(self):
        self.add_action(self.name + '.stack', "Stack", self.stack,
                        tip="Stack selected signals along a new navigation "
                        "axis")

        self.add_action(self.name + '.split', "Split", self.split,
                        tip="Split selected signal along an axis")

        self.add_action(self.name + '.copy', "Copy", self.copy,
                        tip="Make a copy of all selected signals")

        self.add_action(self.name + '.switch', "Switch spaces", self.switch,
                        tip="Switch navigation and signal spaces")

        self.add_action('stats', tr("Statistics"),
                        self.statistics,
                        icon=None,
                        tip=tr("Print the signal statistics to the console."),
                        selection_callback=self.ui.select_signal)

        self.add_action('histogram', tr("Histogram"),
                        self.histogram,
                        icon=None,
                        tip=tr("Plot a histogram of the signal."),
                        selection_callback=self.ui.select_signal)

        self.add_action('add', tr("Add"),
                        self.add,
                        icon='add.svg',
                        tip=tr("Add the currently selected signals"),
                        selection_callback=self.ui.select_signal)

        self.add_action('subtract', tr("Subtract"),
                        self.subtract,
                        icon='subtract.svg',
                        tip=tr("Subtract two signals"),
                        selection_callback=self.ui.select_signal)

        self.add_action('multiply', tr("Multiply"),
                        self.multiply,
                        icon='multiply.svg',
                        tip=tr("Multiply the currently selected signals"),
                        selection_callback=self.ui.select_signal)

        self.add_action('divide', tr("Divide"),
                        self.divide,
                        icon='divide.svg',
                        tip=tr("Divide two signals"),
                        selection_callback=self.ui.select_signal)

        self.add_action('mean', tr("Mean"),
                        self.mean,
                        icon='mean.svg',
                        tip=tr("Take the mean of the current signal"),
                        selection_callback=self.ui.select_signal)

        self.add_action('sum', tr("Sum"),
                        self.sum,
                        icon='sum.svg',
                        tip=tr("Take the sum of the current signal"),
                        selection_callback=self.ui.select_signal)

        self.add_action('max', tr("Maximum"),
                        self.max,
                        icon='max.svg',
                        tip=tr("Take the maximum of the current signal"),
                        selection_callback=self.ui.select_signal)

        self.add_action('min', tr("Minimum"),
                        self.min,
                        icon='min.svg',
                        tip=tr("Take the sum of the current signal"),
                        selection_callback=self.ui.select_signal)

        self.add_action('std', tr("Std.dev."),
                        self.std,
                        icon='stddev.svg',
                        tip=tr("Take the standard deviation of the current "
                               "signal"),
                        selection_callback=self.ui.select_signal)

        self.add_action('var', tr("Variance"),
                        self.var,
                        icon='variance.svg',
                        tip=tr("Take the variances of the current signal"),
                        selection_callback=self.ui.select_signal)

#        max, min, std, var, diff?, derivate?, integrate_simpson, integrate1D,
#        indexmax, valuemax

    def create_menu(self):
        self.add_menuitem('Signal', self.ui.actions[self.name + '.stack'])
        self.add_menuitem('Signal', self.ui.actions[self.name + '.split'])
        self.add_menuitem('Signal', self.ui.actions[self.name + '.copy'])
        self.add_menuitem('Signal', self.ui.actions[self.name + '.switch'])
        self.add_menuitem("Signal", self.ui.actions['stats'])
        self.add_menuitem("Signal", self.ui.actions['histogram'])
        self.add_menuitem("Math", self.ui.actions['add'])
        self.add_menuitem("Math", self.ui.actions['subtract'])
        self.add_menuitem("Math", self.ui.actions['multiply'])
        self.add_menuitem("Math", self.ui.actions['divide'])
        self.add_menuitem("Math", self.ui.actions['mean'])
        self.add_menuitem('Math', self.ui.actions['sum'])
        self.add_menuitem('Math', self.ui.actions['max'])
        self.add_menuitem('Math', self.ui.actions['min'])
        self.add_menuitem('Math', self.ui.actions['std'])
        self.add_menuitem('Math', self.ui.actions['var'])

    def create_toolbars(self):
        self.add_toolbar_button("Math", self.ui.actions['add'])
        self.add_toolbar_button("Math", self.ui.actions['subtract'])
        self.add_toolbar_button("Math", self.ui.actions['multiply'])
        self.add_toolbar_button("Math", self.ui.actions['divide'])
        self.add_toolbar_button("Math", self.ui.actions['mean'])
        self.add_toolbar_button('Math', self.ui.actions['sum'])
        self.add_toolbar_button('Math', self.ui.actions['max'])
        self.add_toolbar_button('Math', self.ui.actions['min'])
        self.add_toolbar_button('Math', self.ui.actions['std'])
        self.add_toolbar_button('Math', self.ui.actions['var'])

    def stack(self, signals=None, advanced=False):
        if signals is None:
            signals = self.ui.get_selected_signals()
        if signals is None:
            raise ValueError("No signals to stack")
        self.record_code('signals = self.ui.get_selected_signals()')
        if advanced:
            # Advanced mode: Pop up a dialog to prompt for axes to sum over=
            axis = self._prompt_axes(signals[0], single=True)
            if not axis:
                return
            stack = stack_signals(signals, axis=axis.name)
            self.record_code('s_stack = hs.stack(signals, axis="%s")' %
                             axis.name)
        else:
            stack = stack_signals(signals)
            self.record_code('s_stack = hs.stack(signals)')
        stack.plot()

    def split(self, signal=None, advanced=False):
        if signal is None:
            signal = self.ui.get_selected_signal()
        if signal is None:
            raise ValueError('No signal to split')
        axis = 'auto'
        if advanced:
            axis = self._prompt_axes(signal, single=True)
            if not axis:
                return
            axis = axis.name
        signals = signal.split(axis=axis)
        for s in signals:
            s.plot()

    def copy(self, signals=None):
        if signals is None:
            signals = self.ui.get_selected_signals()
        for s in signals:
            s.deepcopy().plot()
        self.record_code('signals = self.ui.get_selected_signals()')
        self.record_code('for s in signals:\n\ts.deepcopy().plot()')

    def switch(self, signal=None):
        if signal is None:
            signal = self.ui.get_selected_signal()
        # Switch the nav and sig spaces:
        switched = signal.transpose(signal.axes_manager.navigation_axes,
                                    signal.axes_manager.signal_axes)
        switched.plot()
        self.record_code("signal = ui.get_selected_signal()")
        self.record_code(
            "switched = signal.transpose(" +
            "signal.axes_manager.navigation_axes, " +
            "signal.axes_manager.signal_axes)")

    def statistics(self, signal=None):
        if signal is None:
            signal = self.ui.get_selected_signal()
        # If we're using an external console for errors, we need to temporarily
        # redirect stdout
        _old_stdout = None
        if self.ui.settings['output_to_console', bool]:
            _old_stdout = sys.stdout
            sys.stdout = self.ui.console.kernel.stdout
        try:
            signal.print_summary_statistics()
            self.record_code("signal = ui.get_selected_signal()")
            self.record_code("signal.print_summary_statistics()")
        finally:
            if self.ui.settings['output_to_console', bool]:
                sys.stdout = _old_stdout

    def histogram(self, signal=None):
        signal = signal or self.ui.get_selected_signal()
        if signal is not None:
            method = self.settings['histogram_bins_method']
            signal.get_histogram(bins=method).plot()
            self.record_code("signal = ui.get_selected_signal()")
            self.record_code("histogram = signal.get_histogram(bins=%s)" %
                             method)

    def _prompt_axes(self, signal, single=False):
        diag = AxesPickerDialog(self.ui, signal, single)
        diag.setWindowTitle("Select axes to operate on")
        dr = diag.exec_()
        if dr == QtWidgets.QDialog.Accepted:
            return diag.selected_axes
        return None

    def _np_method(self, name, signal, advanced):
        if signal is None:
            signal = self.ui.get_selected_signal()
        f = getattr(signal, name)
        if advanced:
            # Advanced mode: Pop up a dialog to prompt for axes to sum over=
            axes = self._prompt_axes(signal)
            if not axes:
                return
            f(axis=axes).plot()
        else:
            f().plot()
        self.record_code("signal = ui.get_selected_signal()")
        self.record_code("result = signal.%s()" % name)

    def mean(self, signal=None, advanced=False):
        self._np_method('mean', signal, advanced)

    def sum(self, signal=None, advanced=False):
        self._np_method('sum', signal, advanced)

    def max(self, signal=None, advanced=False):
        self._np_method('max', signal, advanced)

    def min(self, signal=None, advanced=False):
        self._np_method('min', signal, advanced)

    def std(self, signal=None, advanced=False):
        self._np_method('std', signal, advanced)

    def var(self, signal=None, advanced=False):
        self._np_method('var', signal, advanced)

    def add(self, signals=None):
        if signals is None:
            signals = self.ui.get_selected_signals()
        if len(signals) < 2:
            raise ValueError(
                "Two or more signals should be selected for addition!")
        s = signals[0] + signals[1]
        for sb in signals[2:]:
            s += sb
        s.plot()
        self.record_code("signals = ui.get_selected_signals()")
        self.record_code("result = signals[0] + signals[1]")
        if len(signals) == 2:
            self.record_code("for other in signals[2:]:")
            self.record_code("    result += other")

    def subtract(self, signals=None):
        if signals is None:
            signals = self.ui.select_x_signals(2, ["a", " - b"])
            signals = [s.signal for s in signals]
            self.record_code(
                "signals = [s.signal for s in "
                "ui.select_x_signals(2, [\"a\", \" - b\"])")
        elif len(signals != 2):
            raise ValueError(
                "Subtraction can only be performed with two signals")
        result = signals[0] - signals[1]
        result.plot()
        self.record_code("result = signals[0] - signals[1]")

    def multiply(self, signals=None):
        if signals is None:
            signals = self.ui.get_selected_signals()
        if len(signals) < 2:
            raise ValueError(
                "Two or more signals should be selected for multiplication!")
        s = signals[0] * signals[1]
        for sb in signals[2:]:
            s *= sb
        s.plot()
        self.record_code("signals = ui.get_selected_signals()")
        self.record_code("result = signals[0] * signals[1]")
        if len(signals) == 2:
            self.record_code("for other in signals[2:]:")
            self.record_code("    result *= other")

    def divide(self, signals=None):
        if signals is None:
            signals = self.ui.select_x_signals(2, ["a", " - b"])
            signals = [s.signal for s in signals]
            self.record_code(
                "signals = [s.signal for s in "
                "ui.select_x_signals(2, [\"a\", \" - b\"])")
        elif len(signals != 2):
            raise ValueError(
                "Division can only be performed with two signals")
        result = signals[0] / signals[1]
        result.plot()
        self.record_code("result = signals[0] / signals[1]")



import sys
from hyperspyui.singleapplication import get_app
from qtpy import API

key = 'test_single_app'


def handle_second_instance(argv):
    print(argv)
    app.exit()


try:
    app = get_app(key)
except SystemExit:
    print('Exiting early!')
    raise

if "pyqt" in API:
    app.messageAvailable.connect(handle_second_instance)
elif API == 'pyside':
    app.messageReceived.connect(handle_second_instance)

print('Loaded app')
sys.exit(app.exec_())
